"""
PynamoDB lowest level connection
"""

from pynamodb.connection.base import Connection
from pynamodb.connection.table import TableConnection


__all__ = [
    "Connection",
    "TableConnection",
]



from __future__ import annotations
from typing import Any

from typing_extensions import assert_type


def test_model_count() -> None:
    from pynamodb.models import Model
    from pynamodb.expressions.operand import Path

    class MyModel(Model):
        pass

    assert_type(MyModel.count('hash', Path('a').between(1, 3)), int)


def test_model_query() -> None:
    from pynamodb.attributes import NumberAttribute
    from pynamodb.models import Model

    class MyModel(Model):
        my_attr = NumberAttribute()

    # test conditions
    MyModel.query(123, range_key_condition=(MyModel.my_attr == 5), filter_condition=(MyModel.my_attr == 5))

    # test conditions are optional
    MyModel.query(123, range_key_condition=None, filter_condition=None)


def test_pagination() -> None:
    from pynamodb.attributes import NumberAttribute
    from pynamodb.models import Model

    class MyModel(Model):
        my_attr = NumberAttribute()

    result_iterator = MyModel.query(123)
    for model in result_iterator:
        assert_type(model, MyModel)
    if result_iterator.last_evaluated_key:
        assert_type(result_iterator.last_evaluated_key['my_attr'], dict[str, Any])


def test_model_update() -> None:
    from pynamodb.attributes import NumberAttribute, UnicodeAttribute
    from pynamodb.models import Model

    class MyModel(Model):
        my_attr = NumberAttribute()
        my_str_attr = UnicodeAttribute()

    my_model = MyModel()
    my_model.update(actions=[
        # test update expressions
        MyModel.my_attr.set(MyModel.my_attr + 123),
        MyModel.my_attr.set(123 + MyModel.my_attr),
        MyModel.my_attr.set(MyModel.my_attr - 123),
        MyModel.my_attr.set(123 - MyModel.my_attr),
        MyModel.my_attr.set(MyModel.my_attr | 123),
    ])

    _ = MyModel.my_attr.set('foo')  # type:ignore[arg-type]
    _ = MyModel.my_attr.set(MyModel.my_str_attr)  # type:ignore[arg-type]


def test_paths() -> None:
    import pynamodb.expressions.operand
    import pynamodb.expressions.condition
    from pynamodb.attributes import ListAttribute, MapAttribute, UnicodeAttribute
    from pynamodb.models import Model

    class MyMap(MapAttribute):
        my_sub_attr = UnicodeAttribute()

    class MyModel(Model):
        my_list = ListAttribute(of=MyMap)
        my_map = MyMap()

    assert_type(MyModel.my_list[0], pynamodb.expressions.operand.Path)
    assert_type(MyModel.my_list[0] == MyModel(), pynamodb.expressions.condition.Comparison)
    # the following string indexing is not type checked - not by mypy nor in runtime
    assert_type(MyModel.my_list[0]['my_sub_attr'] == 'foobar', pynamodb.expressions.condition.Comparison)
    assert_type(MyModel.my_map == 'foobar', pynamodb.expressions.condition.Comparison)


def test_index_query_scan() -> None:
    from pynamodb.attributes import NumberAttribute
    from pynamodb.models import Model
    from pynamodb.indexes import GlobalSecondaryIndex
    from pynamodb.pagination import ResultIterator

    class UntypedIndex(GlobalSecondaryIndex):
        bar = NumberAttribute(hash_key=True)

    class TypedIndex(GlobalSecondaryIndex['MyModel']):
        bar = NumberAttribute(hash_key=True)

    class MyModel(Model):
        foo = NumberAttribute(hash_key=True)
        bar = NumberAttribute()

        untyped_index = UntypedIndex()
        typed_index = TypedIndex()

    # Ensure old code keeps working
    untyped_query_result: ResultIterator = MyModel.untyped_index.query(123)
    assert_type(next(untyped_query_result), Any)

    # Allow users to specify which model their indices return
    typed_query_result: ResultIterator[MyModel] = MyModel.typed_index.query(123)
    assert_type(next(typed_query_result), MyModel)

    # Ensure old code keeps working
    untyped_scan_result = MyModel.untyped_index.scan()
    assert_type(next(untyped_scan_result), Any)

    # Allow users to specify which model their indices return
    typed_scan_result = MyModel.typed_index.scan()
    assert_type(next(typed_scan_result), MyModel)


def test_map_attribute_derivation() -> None:
    from pynamodb.attributes import MapAttribute

    class MyMap(MapAttribute, object):
        pass


def test_is_in() -> None:
    from pynamodb.models import Model
    from pynamodb.attributes import UnicodeAttribute

    class MyModel(Model):
        attr = UnicodeAttribute()

    _ = MyModel.attr.is_in('foo', 'bar')
    _ = MyModel.attr.is_in(123)  # type:ignore[arg-type]
    _ = MyModel.attr.is_in(['foo', 'bar'])  # type:ignore[arg-type]


def test_append() -> None:
    from pynamodb.models import Model
    from pynamodb.attributes import ListAttribute, NumberAttribute

    class MyModel(Model):
        attr = ListAttribute(of=NumberAttribute)

    MyModel.attr.append(42)  # type:ignore[arg-type]
    MyModel.attr.append([42])
    MyModel.attr.prepend(42)  # type:ignore[arg-type]
    MyModel.attr.prepend([42])



import sys
from typing import Dict
from typing import List

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

if sys.version_info >= (3, 11):
    from typing import NotRequired
else:
    from typing_extensions import NotRequired


class SchemaAttrDefinition(TypedDict):
    AttributeName: str
    AttributeType: str


class KeySchema(TypedDict):
    AttributeName: str
    KeyType: str


class Projection(TypedDict):
    ProjectionType: str
    NonKeyAttributes: NotRequired[List[str]]


class IndexSchema(TypedDict):
    index_name: str
    key_schema: List[Dict[str, str]]
    projection: Dict[str, str]
    attribute_definitions: List[SchemaAttrDefinition]


class ProvisionedThroughput(TypedDict, total=False):
    ReadCapacityUnits: int
    WriteCapacityUnits: int


class GlobalSecondaryIndexSchema(IndexSchema):
    provisioned_throughput: ProvisionedThroughput


class ModelSchema(TypedDict):
    attribute_definitions: List[SchemaAttrDefinition]
    key_schema: List[KeySchema]
    global_secondary_indexes: List[GlobalSecondaryIndexSchema]
    local_secondary_indexes: List[IndexSchema]



import uuid
from datetime import datetime

import botocore.exceptions
import pytest

from pynamodb.connection import Connection
from pynamodb.exceptions import CancellationReason
from pynamodb.exceptions import DoesNotExist, TransactWriteError, InvalidStateError


from pynamodb.attributes import (
    NumberAttribute, UnicodeAttribute, UTCDateTimeAttribute, BooleanAttribute, VersionAttribute
)
from pynamodb.transactions import TransactGet, TransactWrite

from pynamodb.models import Model

IDEMPOTENT_PARAMETER_MISMATCH = 'IdempotentParameterMismatchException'
PROVISIONED_THROUGHPUT_EXCEEDED = 'ProvisionedThroughputExceededException'
RESOURCE_NOT_FOUND = 'ResourceNotFoundException'
TRANSACTION_CANCELLED = 'TransactionCanceledException'
TRANSACTION_IN_PROGRESS = 'TransactionInProgressException'
VALIDATION_EXCEPTION = 'ValidationException'


class User(Model):
    class Meta:
        region = 'us-east-1'
        table_name = 'user'

    user_id = NumberAttribute(hash_key=True)


class BankStatement(Model):

    class Meta:
        region = 'us-east-1'
        table_name = 'statement'

    user_id = NumberAttribute(hash_key=True)
    balance = NumberAttribute(default=0)
    active = BooleanAttribute(default=True)


class LineItem(Model):

    class Meta:
        region = 'us-east-1'
        table_name = 'line-item'

    user_id = NumberAttribute(hash_key=True)
    created_at = UTCDateTimeAttribute(range_key=True, default=datetime.now())
    amount = NumberAttribute()
    currency = UnicodeAttribute()


class DifferentRegion(Model):

    class Meta:
        region = 'us-east-2'
        table_name = 'different-region'

    entry_index = NumberAttribute(hash_key=True)


class Foo(Model):
    class Meta:
        region = 'us-east-1'
        table_name = 'foo'

    bar = NumberAttribute(hash_key=True)
    star = UnicodeAttribute(null=True)
    version = VersionAttribute()


TEST_MODELS = [
    BankStatement,
    DifferentRegion,
    LineItem,
    User,
    Foo
]


@pytest.fixture(scope='module')
def connection(ddb_url):
    yield Connection(host=ddb_url)


@pytest.fixture(scope='module', autouse=True)
def create_tables(ddb_url):
    for m in TEST_MODELS:
        m.Meta.host = ddb_url
        m.create_table(
            read_capacity_units=10,
            write_capacity_units=10,
            wait=True
        )

    yield

    for m in TEST_MODELS:
        if m.exists():
            m.delete_table()


@pytest.mark.ddblocal
def test_transact_write__error__idempotent_parameter_mismatch(connection):
    client_token = str(uuid.uuid4())

    with TransactWrite(connection=connection, client_request_token=client_token) as transaction:
        transaction.save(User(1))
        transaction.save(User(2))

    with pytest.raises(TransactWriteError) as exc_info:
        # committing the first time, then adding more info and committing again
        with TransactWrite(connection=connection, client_request_token=client_token) as transaction:
            transaction.save(User(3))
    assert exc_info.value.cause_response_code == IDEMPOTENT_PARAMETER_MISMATCH
    assert isinstance(exc_info.value.cause, botocore.exceptions.ClientError)
    assert User.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE

    # ensure that the first request succeeded in creating new users
    assert User.get(1)
    assert User.get(2)

    with pytest.raises(DoesNotExist):
        # ensure it did not create the user from second request
        User.get(3)


@pytest.mark.ddblocal
def test_transact_write__error__different_regions(connection):
    # Tip: This test *WILL* fail if run against `dynamodb-local -sharedDb` !
    with pytest.raises(TransactWriteError) as exc_info:
        with TransactWrite(connection=connection) as transact_write:
            # creating a model in a table outside the region everyone else operates in
            transact_write.save(DifferentRegion(entry_index=0))
            transact_write.save(BankStatement(1))
            transact_write.save(User(1))
    assert exc_info.value.cause_response_code == RESOURCE_NOT_FOUND
    assert isinstance(exc_info.value.cause, botocore.exceptions.ClientError)
    assert DifferentRegion.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE
    assert BankStatement.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE
    assert User.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE


@pytest.mark.ddblocal
def test_transact_write__error__transaction_cancelled__condition_check_failure(connection):
    # create a users and a bank statements for them
    User(1).save()
    BankStatement(1).save()

    # attempt to do this as a transaction with the condition that they don't already exist
    with pytest.raises(TransactWriteError) as exc_info:
        with TransactWrite(connection=connection) as transaction:
            transaction.save(User(1), condition=(User.user_id.does_not_exist()))
            transaction.save(BankStatement(1), condition=(BankStatement.user_id.does_not_exist()))
    assert exc_info.value.cause_response_code == TRANSACTION_CANCELLED
    assert 'ConditionalCheckFailed' in exc_info.value.cause_response_message
    assert exc_info.value.cancellation_reasons == [
        CancellationReason(code='ConditionalCheckFailed', message='The conditional request failed'),
        CancellationReason(code='ConditionalCheckFailed', message='The conditional request failed'),
    ]
    assert isinstance(exc_info.value.cause, botocore.exceptions.ClientError)
    assert User.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE
    assert BankStatement.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE


@pytest.mark.ddblocal
def test_transact_write__error__transaction_cancelled__partial_failure(connection):
    User(2).delete()
    BankStatement(2).save()

    # attempt to do this as a transaction with the condition that they don't already exist
    with pytest.raises(TransactWriteError) as exc_info:
        with TransactWrite(connection=connection) as transaction:
            transaction.save(User(2), condition=(User.user_id.does_not_exist()))
            transaction.save(BankStatement(2), condition=(BankStatement.user_id.does_not_exist()))
    assert exc_info.value.cause_response_code == TRANSACTION_CANCELLED
    assert exc_info.value.cancellation_reasons == [
        None,
        CancellationReason(code='ConditionalCheckFailed', message='The conditional request failed'),
    ]


@pytest.mark.ddblocal
def test_transact_write__error__multiple_operations_on_same_record(connection):
    BankStatement(1).save()

    # attempt to do a transaction with multiple operations on the same record
    with pytest.raises(TransactWriteError) as exc_info:
        with TransactWrite(connection=connection) as transaction:
            transaction.condition_check(BankStatement, 1, condition=(BankStatement.user_id.exists()))
            transaction.update(BankStatement(1), actions=[(BankStatement.balance.add(10))])
    assert exc_info.value.cause_response_code == VALIDATION_EXCEPTION
    assert isinstance(exc_info.value.cause, botocore.exceptions.ClientError)
    assert BankStatement.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE


@pytest.mark.ddblocal
def test_transact_get(connection):
    # making sure these entries exist, and with the expected info
    User(1).save()
    BankStatement(1).save()
    User(2).save()
    BankStatement(2, balance=100).save()

    # get users and statements we just created and assign them to variables
    with TransactGet(connection=connection) as transaction:
        _user1_future = transaction.get(User, 1)
        _statement1_future = transaction.get(BankStatement, 1)
        _user2_future = transaction.get(User, 2)
        _statement2_future = transaction.get(BankStatement, 2)

    user1 = _user1_future.get()
    statement1 = _statement1_future.get()
    user2 = _user2_future.get()
    statement2 = _statement2_future.get()

    assert user1.user_id == statement1.user_id == 1
    assert statement1.balance == 0
    assert user2.user_id == statement2.user_id == 2
    assert statement2.balance == 100


@pytest.mark.ddblocal
def test_transact_get__does_not_exist(connection):
    with TransactGet(connection=connection) as transaction:
        _user_future = transaction.get(User, 100)
    with pytest.raises(User.DoesNotExist):
        _user_future.get()


@pytest.mark.ddblocal
def test_transact_get__invalid_state(connection):
    with TransactGet(connection=connection) as transaction:
        _user_future = transaction.get(User, 100)
        with pytest.raises(InvalidStateError):
            _user_future.get()


@pytest.mark.ddblocal
def test_transact_write(connection):
    # making sure these entries exist, and with the expected info
    BankStatement(1, balance=0).save()
    BankStatement(2, balance=100).save()

    # assert values are what we think they should be
    statement1 = BankStatement.get(1)
    statement2 = BankStatement.get(2)
    assert statement1.balance == 0
    assert statement2.balance == 100

    with TransactWrite(connection=connection) as transaction:
        # let the users send money to one another
        # create a credit line item to user 1's account
        transaction.save(
            LineItem(user_id=1, amount=50, currency='USD'),
            condition=(LineItem.user_id.does_not_exist()),
        )
        # create a debit to user 2's account
        transaction.save(
            LineItem(user_id=2, amount=-50, currency='USD'),
            condition=(LineItem.user_id.does_not_exist()),
        )

        # add credit to user 1's account
        transaction.update(statement1, actions=[BankStatement.balance.add(50)])
        # debit from user 2's account if they have enough in the bank
        transaction.update(
            statement2,
            actions=[BankStatement.balance.add(-50)],
            condition=(BankStatement.balance >= 50)
        )

    statement1.refresh()
    statement2.refresh()
    assert statement1.balance == statement2.balance == 50


@pytest.mark.ddblocal
def test_transact_write__one_of_each(connection):
    User(1).save()
    User(2).save()
    statement = BankStatement(1, balance=100, active=True)
    statement.save()

    with TransactWrite(connection=connection) as transaction:
        transaction.condition_check(User, 1, condition=(User.user_id.exists()))
        transaction.delete(User(2))
        transaction.save(LineItem(4, amount=100, currency='USD'), condition=(LineItem.user_id.does_not_exist()))
        transaction.update(
            statement,
            actions=[
                BankStatement.active.set(False),
                BankStatement.balance.set(0),
            ]
        )

    # confirming transaction correct and successful
    assert User.get(1)
    with pytest.raises(DoesNotExist):
        User.get(2)

    new_line_item = next(LineItem.query(4, scan_index_forward=False, limit=1), None)
    assert new_line_item
    assert new_line_item.amount == 100
    assert new_line_item.currency == 'USD'

    statement.refresh()
    assert not statement.active
    assert statement.balance == 0


@pytest.mark.ddblocal
def test_transaction_write_with_version_attribute(connection):
    foo1 = Foo(1)
    foo1.save()
    foo2 = Foo(2, star='bar')
    foo2.save()
    foo3 = Foo(3)
    foo3.save()

    with TransactWrite(connection=connection) as transaction:
        transaction.condition_check(Foo, 1, condition=(Foo.bar.exists()))
        transaction.delete(foo2)
        transaction.save(Foo(4))
        transaction.update(
            foo3,
            actions=[
                Foo.star.set('birdistheword'),
            ]
        )

    assert Foo.get(1).version == 1
    with pytest.raises(DoesNotExist):
        Foo.get(2)
    # Local object's version attribute is updated automatically.
    assert foo3.version == 2
    assert Foo.get(4).version == 1


@pytest.mark.ddblocal
def test_transaction_get_with_version_attribute(connection):
    Foo(11).save()
    Foo(12, star='bar').save()

    with TransactGet(connection=connection) as transaction:
        foo1_future = transaction.get(Foo, 11)
        foo2_future = transaction.get(Foo, 12)

    foo1 = foo1_future.get()
    assert foo1.version == 1
    foo2 = foo2_future.get()
    assert foo2.version == 1
    assert foo2.star == 'bar'


@pytest.mark.ddblocal
def test_transaction_write_with_version_attribute_condition_failure(connection):
    foo = Foo(21)
    foo.save()

    foo2 = Foo(21)

    with pytest.raises(TransactWriteError) as exc_info:
        with TransactWrite(connection=connection) as transaction:
            transaction.save(Foo(21))
    assert exc_info.value.cause_response_code == TRANSACTION_CANCELLED
    assert 'ConditionalCheckFailed' in exc_info.value.cause_response_message
    assert isinstance(exc_info.value.cause, botocore.exceptions.ClientError)
    assert Foo.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE

    with pytest.raises(TransactWriteError) as exc_info:
        with TransactWrite(connection=connection) as transaction:
            transaction.update(
                foo2,
                actions=[
                    Foo.star.set('birdistheword'),
                ]
            )
    assert exc_info.value.cause_response_code == TRANSACTION_CANCELLED
    assert 'ConditionalCheckFailed' in exc_info.value.cause_response_message
    assert Foo.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE
    # Version attribute is not updated on failure.
    assert foo2.version is None

    with pytest.raises(TransactWriteError) as exc_info:
        with TransactWrite(connection=connection) as transaction:
            transaction.delete(foo2)
    assert exc_info.value.cause_response_code == TRANSACTION_CANCELLED
    assert 'ConditionalCheckFailed' in exc_info.value.cause_response_message
    assert Foo.Meta.table_name in exc_info.value.cause.MSG_TEMPLATE



# Copyright (c) 2010-2013 Samuel Sutch [samuel.sutch@gmail.com]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import datetime, time, functools, operator

default_fudge = datetime.timedelta(seconds=0, microseconds=0, days=0)


def deep_eq(_v1, _v2, datetime_fudge=default_fudge, _assert=False):
    """
    Tests for deep equality between two python data structures recursing
    into sub-structures if necessary. Works with all python types including
    iterators and generators. This function was dreampt up to test API responses
    but could be used for anything. Be careful. With deeply nested structures
    you may blow the stack.

    Options:
              datetime_fudge => this is a datetime.timedelta object which, when
                                comparing dates, will accept values that differ
                                by the number of seconds specified
              _assert        => passing yes for this will raise an assertion error
                                when values do not match, instead of returning
                                false (very useful in combination with pdb)

    Doctests included:

    >>> x1, y1 = ({'a': 'b'}, {'a': 'b'})
    >>> deep_eq(x1, y1)
    True
    >>> x2, y2 = ({'a': 'b'}, {'b': 'a'})
    >>> deep_eq(x2, y2)
    False
    >>> x3, y3 = ({'a': {'b': 'c'}}, {'a': {'b': 'c'}})
    >>> deep_eq(x3, y3)
    True
    >>> x4, y4 = ({'c': 't', 'a': {'b': 'c'}}, {'a': {'b': 'n'}, 'c': 't'})
    >>> deep_eq(x4, y4)
    False
    >>> x5, y5 = ({'a': [1,2,3]}, {'a': [1,2,3]})
    >>> deep_eq(x5, y5)
    True
    >>> x6, y6 = ({'a': [1,'b',8]}, {'a': [2,'b',8]})
    >>> deep_eq(x6, y6)
    False
    >>> x7, y7 = ('a', 'a')
    >>> deep_eq(x7, y7)
    True
    >>> x8, y8 = (['p','n',['asdf']], ['p','n',['asdf']])
    >>> deep_eq(x8, y8)
    True
    >>> x9, y9 = (['p','n',['asdf',['omg']]], ['p', 'n', ['asdf',['nowai']]])
    >>> deep_eq(x9, y9)
    False
    >>> x10, y10 = (1, 2)
    >>> deep_eq(x10, y10)
    False
    >>> deep_eq((str(p) for p in xrange(10)), (str(p) for p in xrange(10)))
    True
    >>> str(deep_eq(range(4), range(4)))
    'True'
    >>> deep_eq(xrange(100), xrange(100))
    True
    >>> deep_eq(xrange(2), xrange(5))
    False
    >>> import datetime
    >>> from datetime import datetime as dt
    >>> d1, d2 = (dt.now(), dt.now() + datetime.timedelta(seconds=4))
    >>> deep_eq(d1, d2)
    False
    >>> deep_eq(d1, d2, datetime_fudge=datetime.timedelta(seconds=5))
    True
    """
    _deep_eq = functools.partial(deep_eq, datetime_fudge=datetime_fudge,
                                 _assert=_assert)

    def _check_assert(R, a, b, reason=''):
        if _assert and not R:
            assert 0, "an assertion has failed in deep_eq ({}) {} != {}".format(
                reason, str(a), str(b))
        return R

    def _deep_dict_eq(d1, d2):
        k1, k2 = (sorted(d1.keys()), sorted(d2.keys()))
        if k1 != k2:  # keys should be exactly equal
            return _check_assert(False, k1, k2, "keys")

        return _check_assert(operator.eq(sum(_deep_eq(d1[k], d2[k])
                                             for k in k1),
                                         len(k1)), d1, d2, "dictionaries")

    def _deep_iter_eq(l1, l2):
        if len(l1) != len(l2):
            return _check_assert(False, l1, l2, "lengths")
        return _check_assert(operator.eq(sum(_deep_eq(v1, v2)
                                             for v1, v2 in zip(l1, l2)),
                                         len(l1)), l1, l2, "iterables")

    def op(a, b):
        _op = operator.eq
        if type(a) == datetime.datetime and type(b) == datetime.datetime:
            s = datetime_fudge.seconds
            t1, t2 = (time.mktime(a.timetuple()), time.mktime(b.timetuple()))
            l = t1 - t2
            l = -l if l > 0 else l
            return _check_assert((-s if s > 0 else s) <= l, a, b, "dates")
        return _check_assert(_op(a, b), a, b, "values")

    c1, c2 = (_v1, _v2)

    # guard against strings because they are iterable and their
    # elements yield iterables infinitely.
    # I N C E P T I O N
    if not isinstance(_v1, str):
        if isinstance(_v1, dict):
            op = _deep_dict_eq
        else:
            try:
                c1, c2 = (list(iter(_v1)), list(iter(_v2)))
            except TypeError:
                c1, c2 = _v1, _v2
            else:
                op = _deep_iter_eq

    return op(c1, c2)



import unittest.mock
import pytest

from pynamodb.connection import Connection
from pynamodb.signals import _FakeNamespace
from pynamodb.signals import pre_dynamodb_send, post_dynamodb_send

try:
    import blinker
except ImportError:
    blinker = None

PATCH_METHOD = 'pynamodb.connection.Connection._make_api_call'


@unittest.mock.patch(PATCH_METHOD)
@unittest.mock.patch('pynamodb.connection.base.uuid')
def test_signal(mock_uuid, mock_req):
    pre_recorded = []
    post_recorded = []
    UUID = '123-abc'

    def record_pre_dynamodb_send(sender, operation_name, table_name, req_uuid):
        pre_recorded.append((operation_name, table_name, req_uuid))

    def record_post_dynamodb_send(sender, operation_name, table_name, req_uuid):
        post_recorded.append((operation_name, table_name, req_uuid))

    pre_dynamodb_send.connect(record_pre_dynamodb_send)
    post_dynamodb_send.connect(record_post_dynamodb_send)
    try:
        mock_uuid.uuid4.return_value = UUID
        mock_req.return_value = {'TableDescription': {'TableName': 'table', 'TableStatus': 'Creating'}}
        c = Connection()
        c.dispatch('CreateTable', {'TableName': 'MyTable'})
        assert ('CreateTable', 'MyTable', UUID) == pre_recorded[0]
        assert ('CreateTable', 'MyTable', UUID) == post_recorded[0]
    finally:
        pre_dynamodb_send.disconnect(record_pre_dynamodb_send)
        post_dynamodb_send.disconnect(record_post_dynamodb_send)


@unittest.mock.patch(PATCH_METHOD)
@unittest.mock.patch('pynamodb.connection.base.uuid')
def test_signal_exception_pre_signal(mock_uuid, mock_req):
    post_recorded = []
    UUID = '123-abc'

    def record_pre_dynamodb_send(sender, operation_name, table_name, req_uuid):
        raise ValueError()

    def record_post_dynamodb_send(sender, operation_name, table_name, req_uuid):
        post_recorded.append((operation_name, table_name, req_uuid))

    pre_dynamodb_send.connect(record_pre_dynamodb_send)
    post_dynamodb_send.connect(record_post_dynamodb_send)
    try:
        mock_uuid.uuid4.return_value = UUID
        mock_req.return_value = {'TableDescription': {'TableName': 'table', 'TableStatus': 'Creating'}}
        c = Connection()
        c.dispatch('CreateTable', {'TableName': 'MyTable'})
        assert ('CreateTable', 'MyTable', UUID) == post_recorded[0]
    finally:
        pre_dynamodb_send.disconnect(record_pre_dynamodb_send)
        post_dynamodb_send.disconnect(record_post_dynamodb_send)


@unittest.mock.patch(PATCH_METHOD)
@unittest.mock.patch('pynamodb.connection.base.uuid')
def test_signal_exception_post_signal(mock_uuid, mock_req):
    pre_recorded = []
    UUID = '123-abc'

    def record_pre_dynamodb_send(sender, operation_name, table_name, req_uuid):
        pre_recorded.append((operation_name, table_name, req_uuid))

    def record_post_dynamodb_send(sender, operation_name, table_name, req_uuid):
        raise ValueError()

    pre_dynamodb_send.connect(record_pre_dynamodb_send)
    post_dynamodb_send.connect(record_post_dynamodb_send)
    try:
        mock_uuid.uuid4.return_value = UUID
        mock_req.return_value = {'TableDescription': {'TableName': 'table', 'TableStatus': 'Creating'}}
        c = Connection()
        c.dispatch('CreateTable', {'TableName': 'MyTable'})
        assert ('CreateTable', 'MyTable', UUID) == pre_recorded[0]
    finally:
        pre_dynamodb_send.disconnect(record_pre_dynamodb_send)
        post_dynamodb_send.disconnect(record_post_dynamodb_send)


def test_fake_signals():
    _signals = _FakeNamespace()
    pre_dynamodb_send = _signals.signal('pre_dynamodb_send')
    with pytest.raises(RuntimeError):
        pre_dynamodb_send.connect(lambda x: x)
    pre_dynamodb_send.send(object, operation_name="UPDATE", table_name="TEST", req_uuid="something")



"""
Test suite for the table class
"""
from unittest import TestCase

from pynamodb.connection import TableConnection
from pynamodb.connection.base import MetaTable
from pynamodb.constants import TABLE_KEY
from pynamodb.expressions.operand import Path
from .data import DESCRIBE_TABLE_DATA, GET_ITEM_DATA
from .response import HttpOK

from unittest.mock import patch

PATCH_METHOD = 'pynamodb.connection.Connection._make_api_call'


class ConnectionTestCase(TestCase):
    """
    Tests for the base connection class
    """

    def setUp(self):
        self.test_table_name = 'Thread'
        self.region = 'us-east-1'

    def test_create_connection(self):
        """
        TableConnection()
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
        self.assertIsNotNone(conn)

    def test_connection_session_set_credentials(self):
        conn = TableConnection(
            self.test_table_name,
            meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]),
            aws_access_key_id='access_key_id',
            aws_secret_access_key='secret_access_key')

        credentials = conn.connection.session.get_credentials()

        self.assertEqual(credentials.access_key, 'access_key_id')
        self.assertEqual(credentials.secret_key, 'secret_access_key')

    def test_connection_session_set_credentials_with_session_token(self):
        conn = TableConnection(
            self.test_table_name,
            meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]),
            aws_access_key_id='access_key_id',
            aws_secret_access_key='secret_access_key',
            aws_session_token='session_token')

        credentials = conn.connection.session.get_credentials()

        self.assertEqual(credentials.access_key, 'access_key_id')
        self.assertEqual(credentials.secret_key, 'secret_access_key')
        self.assertEqual(credentials.token, 'session_token')

    def test_create_table(self):
        """
        TableConnection.create_table
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
        kwargs = {
            'read_capacity_units': 1,
            'write_capacity_units': 1,
        }
        self.assertRaises(ValueError, conn.create_table, **kwargs)
        kwargs['attribute_definitions'] = [
            {
                'attribute_name': 'key1',
                'attribute_type': 'S'
            },
            {
                'attribute_name': 'key2',
                'attribute_type': 'S'
            }
        ]
        self.assertRaises(ValueError, conn.create_table, **kwargs)
        kwargs['key_schema'] = [
            {
                'attribute_name': 'key1',
                'key_type': 'hash'
            },
            {
                'attribute_name': 'key2',
                'key_type': 'range'
            }
        ]
        params = {
            'TableName': 'Thread',
            'ProvisionedThroughput': {
                'WriteCapacityUnits': 1,
                'ReadCapacityUnits': 1
            },
            'AttributeDefinitions': [
                {
                    'AttributeType': 'S',
                    'AttributeName': 'key1'
                },
                {
                    'AttributeType': 'S',
                    'AttributeName': 'key2'
                }
            ],
            'KeySchema': [
                {
                    'KeyType': 'HASH',
                    'AttributeName': 'key1'
                },
                {
                    'KeyType': 'RANGE',
                    'AttributeName': 'key2'
                }
            ]
        }
        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.create_table(
                **kwargs
            )
            kwargs = req.call_args[0][1]
            self.assertEqual(kwargs, params)

    def test_create_table_with_tags(self):
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
        kwargs = {
            'read_capacity_units': 1,
            'write_capacity_units': 1,
            'attribute_definitions': [
                {
                    'attribute_name': 'key1',
                    'attribute_type': 'S'
                },
                {
                    'attribute_name': 'key2',
                    'attribute_type': 'S'
                }
            ],
            'key_schema': [
                {
                    'attribute_name': 'key1',
                    'key_type': 'hash'
                },
                {
                    'attribute_name': 'key2',
                    'key_type': 'range'
                }
            ],
            'tags': {
                'tag-key1': 'tag-value1',
                'tag-key2': 'tag-value2',
            }
        }
        params = {
            'TableName': 'Thread',
            'ProvisionedThroughput': {
                'WriteCapacityUnits': 1,
                'ReadCapacityUnits': 1
            },
            'AttributeDefinitions': [
                {
                    'AttributeType': 'S',
                    'AttributeName': 'key1'
                },
                {
                    'AttributeType': 'S',
                    'AttributeName': 'key2'
                }
            ],
            'KeySchema': [
                {
                    'KeyType': 'HASH',
                    'AttributeName': 'key1'
                },
                {
                    'KeyType': 'RANGE',
                    'AttributeName': 'key2'
                }
            ],
            'Tags': [
                {
                    'Key': 'tag-key1',
                    'Value': 'tag-value1'
                },
                {
                    'Key': 'tag-key2',
                    'Value': 'tag-value2'
                }
            ]
        }
        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.create_table(
                **kwargs
            )
            kwargs = req.call_args[0][1]
            self.assertEqual(kwargs, params)

    def test_update_time_to_live(self):
        """
        TableConnection.update_time_to_live
        """
        params = {
            'TableName': 'Thread',
            'TimeToLiveSpecification': {
                'AttributeName': 'ttl_attr',
                'Enabled': True,
            }
        }
        with patch(PATCH_METHOD) as req:
            req.return_value = HttpOK(), None
            conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
            conn.update_time_to_live('ttl_attr')
            kwargs = req.call_args[0][1]
            self.assertEqual(kwargs, params)

    def test_delete_table(self):
        """
        TableConnection.delete_table
        """
        params = {'TableName': 'Thread'}
        with patch(PATCH_METHOD) as req:
            req.return_value = HttpOK(), None
            conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
            conn.delete_table()
            kwargs = req.call_args[0][1]
            self.assertEqual(kwargs, params)

    def test_update_table(self):
        """
        TableConnection.update_table
        """
        with patch(PATCH_METHOD) as req:
            req.return_value = HttpOK(), None
            conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
            params = {
                'ProvisionedThroughput': {
                    'WriteCapacityUnits': 2,
                    'ReadCapacityUnits': 2
                },
                'TableName': self.test_table_name
            }
            conn.update_table(
                read_capacity_units=2,
                write_capacity_units=2
            )
            self.assertEqual(req.call_args[0][1], params)

        with patch(PATCH_METHOD) as req:
            req.return_value = HttpOK(), None
            conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))

            global_secondary_index_updates = [
                {
                    "index_name": "foo-index",
                    "read_capacity_units": 2,
                    "write_capacity_units": 2
                }
            ]
            params = {
                'TableName': self.test_table_name,
                'ProvisionedThroughput': {
                    'ReadCapacityUnits': 2,
                    'WriteCapacityUnits': 2,
                },
                'GlobalSecondaryIndexUpdates': [
                    {
                        'Update': {
                            'IndexName': 'foo-index',
                            'ProvisionedThroughput': {
                                'ReadCapacityUnits': 2,
                                'WriteCapacityUnits': 2,
                            }
                        }
                    }

                ]
            }
            conn.update_table(
                read_capacity_units=2,
                write_capacity_units=2,
                global_secondary_index_updates=global_secondary_index_updates
            )
            self.assertEqual(req.call_args[0][1], params)

    def test_describe_table(self):
        """
        TableConnection.describe_table
        """
        with patch(PATCH_METHOD) as req:
            req.return_value = DESCRIBE_TABLE_DATA
            conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
            data = conn.describe_table()
            self.assertEqual(data, DESCRIBE_TABLE_DATA[TABLE_KEY])
            self.assertEqual(req.call_args[0][1], {'TableName': 'Thread'})

    def test_delete_item(self):
        """
        TableConnection.delete_item
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))

        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.delete_item(
                "Amazon DynamoDB",
                "How do I update multiple items?")
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'Key': {
                    'ForumName': {
                        'S': 'Amazon DynamoDB'
                    },
                    'Subject': {
                        'S': 'How do I update multiple items?'
                    }
                },
                'TableName': self.test_table_name
            }
            self.assertEqual(req.call_args[0][1], params)

    def test_update_item(self):
        """
        TableConnection.update_item
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))

        with patch(PATCH_METHOD) as req:
            req.return_value = HttpOK(), {}
            conn.update_item(
                'foo-key',
                actions=[Path('Subject').set('foo-subject')],
                range_key='foo-range-key',
            )
            params = {
                'Key': {
                    'ForumName': {
                        'S': 'foo-key'
                    },
                    'Subject': {
                        'S': 'foo-range-key'
                    }
                },
                'UpdateExpression': 'SET #0 = :0',
                'ExpressionAttributeNames': {
                    '#0': 'Subject'
                },
                'ExpressionAttributeValues': {
                    ':0': {
                        'S': 'foo-subject'
                    }
                },
                'ReturnConsumedCapacity': 'TOTAL',
                'TableName': 'Thread'
            }
            self.assertEqual(req.call_args[0][1], params)

    def test_get_item(self):
        """
        TableConnection.get_item
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))

        with patch(PATCH_METHOD) as req:
            req.return_value = GET_ITEM_DATA
            item = conn.get_item("Amazon DynamoDB", "How do I update multiple items?")
            self.assertEqual(item, GET_ITEM_DATA)

    def test_put_item(self):
        """
        TableConnection.put_item
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))

        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.put_item(
                'foo-key',
                range_key='foo-range-key',
                attributes={'ForumName': 'foo-value'}
            )
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'TableName': self.test_table_name,
                'Item': {'ForumName': {'S': 'foo-value'}, 'Subject': {'S': 'foo-range-key'}}
            }
            self.assertEqual(req.call_args[0][1], params)

        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.put_item(
                'foo-key',
                range_key='foo-range-key',
                attributes={'ForumName': 'foo-value'}
            )
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'Item': {
                    'ForumName': {
                        'S': 'foo-value'
                    },
                    'Subject': {
                        'S': 'foo-range-key'
                    }
                },
                'TableName': self.test_table_name
            }
            self.assertEqual(req.call_args[0][1], params)

        with patch(PATCH_METHOD) as req:
            req.return_value = HttpOK(), {}
            conn.put_item(
                'foo-key',
                range_key='foo-range-key',
                attributes={'ForumName': 'foo-value'},
                condition=Path('ForumName').does_not_exist()
            )
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'Item': {
                    'ForumName': {
                        'S': 'foo-value'
                    },
                    'Subject': {
                        'S': 'foo-range-key'
                    }
                },
                'TableName': self.test_table_name,
                'ConditionExpression': 'attribute_not_exists (#0)',
                'ExpressionAttributeNames': {
                    '#0': 'ForumName'
                }
            }
            self.assertEqual(req.call_args[0][1], params)

    def test_batch_write_item(self):
        """
        TableConnection.batch_write_item
        """
        items = []
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
        for i in range(10):
            items.append(
                {"ForumName": "FooForum", "Subject": "thread-{}".format(i)}
            )
        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.batch_write_item(
                put_items=items
            )
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'RequestItems': {
                    self.test_table_name: [
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-0'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-1'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-2'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-3'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-4'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-5'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-6'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-7'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-8'}}}},
                        {'PutRequest': {'Item': {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-9'}}}}
                    ]
                }
            }
            self.assertEqual(req.call_args[0][1], params)

    def test_batch_get_item(self):
        """
        TableConnection.batch_get_item
        """
        items = []
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
        for i in range(10):
            items.append(
                {"ForumName": "FooForum", "Subject": "thread-{}".format(i)}
            )

        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.batch_get_item(
                items
            )
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'RequestItems': {
                    self.test_table_name: {
                        'Keys': [
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-0'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-1'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-2'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-3'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-4'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-5'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-6'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-7'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-8'}},
                            {'ForumName': {'S': 'FooForum'}, 'Subject': {'S': 'thread-9'}}
                        ]
                    }
                }
            }
            self.assertEqual(req.call_args[0][1], params)

    def test_query(self):
        """
        TableConnection.query
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))

        with patch(PATCH_METHOD) as req:
            req.return_value = {}
            conn.query(
                "FooForum",
                Path('Subject').startswith('thread')
            )
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'KeyConditionExpression': '(#0 = :0 AND begins_with (#1, :1))',
                'ExpressionAttributeNames': {
                    '#0': 'ForumName',
                    '#1': 'Subject'
                },
                'ExpressionAttributeValues': {
                    ':0': {
                        'S': 'FooForum'
                    },
                    ':1': {
                        'S': 'thread'
                    }
                },
                'TableName': self.test_table_name
            }
            self.assertEqual(req.call_args[0][1], params)

    def test_scan(self):
        """
        TableConnection.scan
        """
        conn = TableConnection(self.test_table_name, meta_table=MetaTable(DESCRIBE_TABLE_DATA[TABLE_KEY]))
        with patch(PATCH_METHOD) as req:
            req.return_value = HttpOK(), {}
            conn.scan()
            params = {
                'ReturnConsumedCapacity': 'TOTAL',
                'TableName': self.test_table_name
            }
            self.assertEqual(req.call_args[0][1], params)



from botocore.exceptions import ClientError

from pynamodb.exceptions import PynamoDBException, PutError


def test_get_cause_response_code():
    error = PutError(
        cause=ClientError(
            error_response={
                'Error': {
                    'Code': 'hello'
                }
            },
            operation_name='test'
        )
    )
    assert error.cause_response_code == 'hello'


def test_get_cause_response_code__no_code():
    error = PutError()
    assert error.cause_response_code is None


def test_get_cause_response_message():
    error = PutError(
        cause=ClientError(
            error_response={
                'Error': {
                    'Message': 'hiya'
                }
            },
            operation_name='test'
        )
    )
    assert error.cause_response_message == 'hiya'


def test_get_cause_response_message__no_message():
    error = PutError()
    assert error.cause_response_message is None


class PynamoDBTestError(PynamoDBException):
    msg = "Test message"


def test_subclass_message_is_not_overwritten_with_none():
    assert PynamoDBTestError().msg == "Test message"



import pytest

import pynamodb.exceptions
from pynamodb.attributes import DiscriminatorAttribute
from pynamodb.attributes import DynamicMapAttribute
from pynamodb.attributes import ListAttribute
from pynamodb.attributes import MapAttribute
from pynamodb.attributes import NumberAttribute
from pynamodb.attributes import UnicodeAttribute
from pynamodb.indexes import AllProjection
from pynamodb.models import Model
from pynamodb.indexes import GlobalSecondaryIndex


class TestDiscriminatorIndex:

    @pytest.mark.ddblocal
    def test_create_table(self, ddb_url):
        class ParentModel(Model, discriminator='Parent'):
            class Meta:
                host = ddb_url
                table_name = 'discriminator_index_test'
                read_capacity_units = 1
                write_capacity_units = 1

            hash_key = UnicodeAttribute(hash_key=True)
            cls = DiscriminatorAttribute()

        class ChildIndex(GlobalSecondaryIndex):
            class Meta:
                index_name = 'child_index'
                projection = AllProjection()
                read_capacity_units = 1
                write_capacity_units = 1

            index_key = UnicodeAttribute(hash_key=True)

        class ChildModel1(ParentModel, discriminator='Child1'):
            child_index = ChildIndex()
            index_key = UnicodeAttribute()

        # Multiple child models can share the same index
        class ChildModel2(ParentModel, discriminator='Child2'):
            child_index = ChildIndex()
            index_key = UnicodeAttribute()

        # What's important to notice is that the child_index is not defined on the parent class.
        # We're running `create_table` on the ParentModel, and expect it to know about child models
        # (through the discriminator association) and include all child models' indexes
        # during table creation.
        ParentModel.create_table(read_capacity_units=1, write_capacity_units=1, wait=True)

        model = ChildModel1()
        model.hash_key = 'hash_key1'
        model.index_key = 'bar'
        model.save()

        model = ChildModel2()
        model.hash_key = 'hash_key2'
        model.index_key = 'baz'
        model.save()

        model = next(ChildModel1.child_index.query('bar'))
        assert isinstance(model, ChildModel1)

        model = next(ChildModel2.child_index.query('baz'))
        assert isinstance(model, ChildModel2)

    @pytest.mark.ddblocal
    def test_create_table__incompatible_indexes(self, ddb_url):
        class ParentModel(Model, discriminator='Parent'):
            class Meta:
                host = ddb_url
                table_name = 'discriminator_index_test__incompatible_indexes'
                read_capacity_units = 1
                write_capacity_units = 1

            hash_key = UnicodeAttribute(hash_key=True)
            cls = DiscriminatorAttribute()

        class ChildIndex1(GlobalSecondaryIndex):
            class Meta:
                index_name = 'child_index1'
                projection = AllProjection()
                read_capacity_units = 1
                write_capacity_units = 1

            index_key = UnicodeAttribute(hash_key=True)

        class ChildIndex2(GlobalSecondaryIndex):
            class Meta:
                index_name = 'child_index2'
                projection = AllProjection()
                read_capacity_units = 1
                write_capacity_units = 1

            # Intentionally a different type from ChildIndex1.index_key
            index_key = NumberAttribute(hash_key=True)

        # noinspection PyUnusedLocal
        class ChildModel1(ParentModel, discriminator='Child1'):
            child_index = ChildIndex1()
            index_key = UnicodeAttribute()

        # noinspection PyUnusedLocal
        class ChildModel2(ParentModel, discriminator='Child2'):
            child_index = ChildIndex2()
            index_key = UnicodeAttribute()

        # Unlike `test_create_table`, we expect this to fail because the child indexes
        # attempt to use the same attribute name for different types, thus the resulting table's
        # AttributeDefinitions would have the same attribute appear twice with conflicting types.
        with pytest.raises(pynamodb.exceptions.TableError, match="Cannot have two attributes with the same name"):
            ParentModel.create_table(read_capacity_units=1, write_capacity_units=1, wait=True)



"""
Type-annotates the private botocore APIs that we're currently relying on.
"""
from typing import Any, Dict, Optional

import botocore.client
import botocore.credentials
import botocore.endpoint
import botocore.hooks
import botocore.model
import botocore.signers


class BotocoreEndpointPrivate(botocore.endpoint.Endpoint):
    _event_emitter: botocore.hooks.HierarchicalEmitter


class BotocoreRequestSignerPrivate(botocore.signers.RequestSigner):
    _credentials: botocore.credentials.Credentials


class BotocoreBaseClientPrivate(botocore.client.BaseClient):
    _endpoint: BotocoreEndpointPrivate
    _request_signer: BotocoreRequestSignerPrivate
    _service_model: botocore.model.ServiceModel

    def _resolve_endpoint_ruleset(
        self,
        operation_model: botocore.model.OperationModel,
        params: Dict[str, Any],
        request_context: Dict[str, Any],
        ignore_signing_region: bool = ...,
    ):
        raise NotImplementedError

    def _convert_to_request_dict(
        self,
        api_params: Dict[str, Any],
        operation_model: botocore.model.OperationModel,
        *,
        endpoint_url: str = ...,  # added in botocore 1.28
        context: Optional[Dict[str, Any]] = ...,
        headers: Optional[Dict[str, Any]] = ...,
        set_user_agent_header: bool = ...,
    ) -> Dict[str, Any]:
        raise NotImplementedError



import time
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, TypeVar

from pynamodb.constants import (CAMEL_COUNT, ITEMS, LAST_EVALUATED_KEY, SCANNED_COUNT,
                                CONSUMED_CAPACITY, TOTAL, CAPACITY_UNITS)
from pynamodb.settings import OperationSettings

_T = TypeVar('_T')


class RateLimiter:
    """
    RateLimiter limits operations to a pre-set rate of units/seconds

    Example:
        Initialize a RateLimiter with the desired rate
            rate_limiter = RateLimiter(rate_limit)

        Now, every time before calling an operation, call acquire()
            rate_limiter.acquire()

        And after an operation, update the number of units consumed
            rate_limiter.consume(units)

    """
    def __init__(self, rate_limit: float, time_module: Optional[Any] = None) -> None:
        """
        Initializes a RateLimiter object

        :param rate_limit: The desired rate
        :param time_module: Optional: the module responsible for calculating time. Intended to be used for testing purposes.
        """
        if rate_limit <= 0:
            raise ValueError("rate_limit must be greater than zero")
        self._rate_limit = rate_limit
        self._consumed = 0
        self._time_of_last_acquire = 0.0
        self._time_module: Any = time_module or time

    def consume(self, units: int) -> None:
        """
        Records the amount of units consumed.

        :param units: Number of units consumed

        :return: None
        """
        self._consumed += units

    def acquire(self) -> None:
        """
        Sleeps the appropriate amount of time to follow the rate limit restriction

        :return: None
        """

        self._time_module.sleep(max(0, self._consumed/float(self.rate_limit) - (self._time_module.time()-self._time_of_last_acquire)))
        self._consumed = 0
        self._time_of_last_acquire = self._time_module.time()

    @property
    def rate_limit(self) -> float:
        """
        A limit of units per seconds
        """
        return self._rate_limit

    @rate_limit.setter
    def rate_limit(self, rate_limit: float):
        if rate_limit <= 0:
            raise ValueError("rate_limit must be greater than zero")
        self._rate_limit = rate_limit


class PageIterator(Iterator[_T]):
    """
    PageIterator handles Query and Scan result pagination.

    https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Query.Pagination.html
    https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Scan.html#Scan.Pagination
    """
    def __init__(
        self,
        operation: Callable,
        args: Any,
        kwargs: Dict[str, Any],
        rate_limit: Optional[float] = None,
        settings: OperationSettings = OperationSettings.default,
    ) -> None:
        self._operation = operation
        self._args = args
        self._kwargs = kwargs
        self._last_evaluated_key = kwargs.get('exclusive_start_key')
        self._is_last_page = False
        self._total_scanned_count = 0
        self._rate_limiter = None
        if rate_limit:
            self._rate_limiter = RateLimiter(rate_limit)
        self._settings = settings

    def __iter__(self) -> Iterator[_T]:
        return self

    def __next__(self) -> _T:
        if self._is_last_page:
            raise StopIteration()

        self._kwargs['exclusive_start_key'] = self._last_evaluated_key

        if self._rate_limiter:
            self._rate_limiter.acquire()
            self._kwargs['return_consumed_capacity'] = TOTAL
        page = self._operation(*self._args, settings=self._settings, **self._kwargs)
        self._last_evaluated_key = page.get(LAST_EVALUATED_KEY)
        self._is_last_page = self._last_evaluated_key is None
        self._total_scanned_count += page[SCANNED_COUNT]

        if self._rate_limiter:
            consumed_capacity = page.get(CONSUMED_CAPACITY, {}).get(CAPACITY_UNITS, 0)
            self._rate_limiter.consume(consumed_capacity)

        return page

    def next(self) -> _T:
        return self.__next__()

    @property
    def key_names(self) -> Iterable[str]:
        # If the current page has a last_evaluated_key, use it to determine key attributes
        if self._last_evaluated_key:
            return self._last_evaluated_key.keys()

        # Use the table meta data to determine the key attributes
        table_meta = self._operation.__self__.get_meta_table()  # type: ignore
        return table_meta.get_key_names(self._kwargs.get('index_name'))

    @property
    def page_size(self) -> Optional[int]:
        return self._kwargs.get('limit')

    @page_size.setter
    def page_size(self, page_size: int) -> None:
        self._kwargs['limit'] = page_size

    @property
    def last_evaluated_key(self) -> Optional[Dict[str, Dict[str, Any]]]:
        return self._last_evaluated_key

    @property
    def total_scanned_count(self) -> int:
        return self._total_scanned_count


class ResultIterator(Iterator[_T]):
    """
    ResultIterator handles Query and Scan item pagination.

    https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Query.Pagination.html
    https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Scan.html#Scan.Pagination
    """
    def __init__(
        self,
        operation: Callable,
        args: Any,
        kwargs: Dict[str, Any],
        map_fn: Optional[Callable] = None,
        limit: Optional[int] = None,
        rate_limit: Optional[float] = None,
        settings: OperationSettings = OperationSettings.default,
    ) -> None:
        self.page_iter: PageIterator = PageIterator(operation, args, kwargs, rate_limit, settings)
        self._map_fn = map_fn
        self._limit = limit
        self._total_count = 0
        self._index = 0
        self._count = 0

    def _get_next_page(self) -> None:
        page = next(self.page_iter)
        self._count = page[CAMEL_COUNT]
        self._items = page.get(ITEMS)  # not returned if 'Select' is set to 'COUNT'
        self._index = 0 if self._items else self._count
        self._total_count += self._count

    def __iter__(self) -> Iterator[_T]:
        return self

    def __next__(self) -> _T:
        if self._limit == 0:
            raise StopIteration

        while self._index == self._count:
            self._get_next_page()

        item = self._items[self._index]
        self._index += 1
        if self._limit is not None:
            self._limit -= 1
        if self._map_fn:
            item = self._map_fn(item)
        return item

    def next(self) -> _T:
        return self.__next__()

    @property
    def last_evaluated_key(self) -> Optional[Dict[str, Dict[str, Any]]]:
        if self._index == self._count:
            # Not started iterating yet: return `exclusive_start_key` if set, otherwise expect None; or,
            # Entire page has been consumed: last_evaluated_key is whatever DynamoDB returned
            # It may correspond to the current item, or it may correspond to an item evaluated but not returned.
            return self.page_iter.last_evaluated_key

        # In the middle of a page of results: reconstruct a last_evaluated_key from the current item
        # The operation should be resumed starting at the last item returned, not the last item evaluated.
        # This can occur if the 'limit' is reached in the middle of a page.
        item = self._items[self._index - 1]
        return {key: item[key] for key in self.page_iter.key_names}

    @property
    def total_count(self) -> int:
        return self._total_count



from typing import Tuple, TypeVar, Type, Any, List, Optional, Dict, Union, Text, Generic

from pynamodb.connection import Connection
from pynamodb.constants import ITEM, RESPONSES
from pynamodb.expressions.condition import Condition
from pynamodb.expressions.update import Action
from pynamodb.models import Model, _ModelFuture, _KeyType

_M = TypeVar('_M', bound=Model)
_TTransaction = TypeVar('_TTransaction', bound='Transaction')


class Transaction:

    """
    Base class for a type of transaction operation
    """

    def __init__(self, connection: Connection, return_consumed_capacity: Optional[str] = None) -> None:
        self._connection = connection
        self._return_consumed_capacity = return_consumed_capacity

    def _commit(self):
        raise NotImplementedError()

    def __enter__(self: _TTransaction) -> _TTransaction:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and exc_val is None and exc_tb is None:
            self._commit()


class TransactGet(Transaction):

    _results: Optional[List] = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._get_items: List[Dict] = []
        self._futures: List[_ModelFuture] = []
        super(TransactGet, self).__init__(*args, **kwargs)

    def get(self, model_cls: Type[_M], hash_key: _KeyType, range_key: Optional[_KeyType] = None) -> _ModelFuture[_M]:
        """
        Adds the operation arguments for an item to list of models to get
        returns a _ModelFuture object as a placeholder

        :param model_cls:
        :param hash_key:
        :param range_key:
        :return:
        """
        operation_kwargs = model_cls.get_operation_kwargs_from_class(hash_key, range_key=range_key)
        model_future = _ModelFuture(model_cls)
        self._futures.append(model_future)
        self._get_items.append(operation_kwargs)
        return model_future

    @staticmethod
    def _update_futures(futures: List[_ModelFuture], results: List) -> None:
        for model, data in zip(futures, results):
            model.update_with_raw_data(data.get(ITEM))

    def _commit(self) -> Any:
        response = self._connection.transact_get_items(
            get_items=self._get_items,
            return_consumed_capacity=self._return_consumed_capacity
        )

        results = response[RESPONSES]
        self._results = results
        self._update_futures(self._futures, results)
        return response


class TransactWrite(Transaction):

    def __init__(
        self,
        client_request_token: Optional[str] = None,
        return_item_collection_metrics: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super(TransactWrite, self).__init__(**kwargs)
        self._client_request_token: Optional[str] = client_request_token
        self._return_item_collection_metrics = return_item_collection_metrics
        self._condition_check_items: List[Dict] = []
        self._delete_items: List[Dict] = []
        self._put_items: List[Dict] = []
        self._update_items: List[Dict] = []
        self._models_for_version_attribute_update: List[Any] = []

    def condition_check(self, model_cls: Type[_M], hash_key: _KeyType, range_key: Optional[_KeyType] = None, condition: Optional[Condition] = None):
        if condition is None:
            raise TypeError('`condition` cannot be None')
        operation_kwargs = model_cls.get_operation_kwargs_from_class(
            hash_key,
            range_key=range_key,
            condition=condition
        )
        self._condition_check_items.append(operation_kwargs)

    def delete(self, model: _M, condition: Optional[Condition] = None) -> None:
        operation_kwargs = model.get_delete_kwargs_from_instance(condition=condition)
        self._delete_items.append(operation_kwargs)

    def save(self, model: _M, condition: Optional[Condition] = None, return_values: Optional[str] = None) -> None:
        operation_kwargs = model.get_save_kwargs_from_instance(
            condition=condition,
            return_values_on_condition_failure=return_values
        )
        self._put_items.append(operation_kwargs)
        self._models_for_version_attribute_update.append(model)

    def update(self, model: _M, actions: List[Action], condition: Optional[Condition] = None, return_values: Optional[str] = None) -> None:
        operation_kwargs = model.get_update_kwargs_from_instance(
            actions=actions,
            condition=condition,
            return_values_on_condition_failure=return_values
        )
        self._update_items.append(operation_kwargs)
        self._models_for_version_attribute_update.append(model)

    def _commit(self) -> Any:
        response = self._connection.transact_write_items(
            condition_check_items=self._condition_check_items,
            delete_items=self._delete_items,
            put_items=self._put_items,
            update_items=self._update_items,
            client_request_token=self._client_request_token,
            return_consumed_capacity=self._return_consumed_capacity,
            return_item_collection_metrics=self._return_item_collection_metrics,
        )
        for model in self._models_for_version_attribute_update:
            model.update_local_version_attribute()
        return response



"""
Implements signals based on blinker if available, otherwise
falls silently back to a noop.

This implementation was taken from Flask:
https://github.com/pallets/flask/blob/master/flask/signals.py
"""
signals_available = False


class _FakeNamespace(object):
    def signal(self, name, doc=None):
        return _FakeSignal(name, doc)


class _FakeSignal(object):
    """
    If blinker is unavailable, create a fake class with the same
    interface that allows sending of signals but will fail with an
    error on anything else.  Instead of doing anything on send, it
    will just ignore the arguments and do nothing instead.
    """

    def __init__(self, name, doc=None):
        self.name = name
        self.__doc__ = doc

    def _fail(self, *args, **kwargs):
        raise RuntimeError('signalling support is unavailable '
                           'because the blinker library is '
                           'not installed.')

    send = lambda *a, **kw: None  # noqa
    connect = disconnect = has_receivers_for = receivers_for = \
        temporarily_connected_to = _fail
    del _fail


try:
    from blinker import Namespace
    signals_available = True
except ImportError: # pragma: no cover
    Namespace = _FakeNamespace

# The namespace for code signals.  If you are not PynamoDB code, do
# not put signals in here.  Create your own namespace instead.
_signals = Namespace()

pre_dynamodb_send = _signals.signal('pre_dynamodb_send')
post_dynamodb_send = _signals.signal('post_dynamodb_send')



"""
pynamodb attributes tests
"""
import calendar
import json

from base64 import b64encode
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import List
from typing import Set

from unittest.mock import patch, call
import pytest

from pynamodb.attributes import (
    BinarySetAttribute, BinaryAttribute, DynamicMapAttribute, NumberSetAttribute, NumberAttribute,
    UnicodeAttribute, UnicodeSetAttribute, UTCDateTimeAttribute, BooleanAttribute, MapAttribute, NullAttribute,
    ListAttribute, JSONAttribute, TTLAttribute, VersionAttribute, Attribute)
from pynamodb.constants import (
    NUMBER, STRING, STRING_SET, NUMBER_SET, BINARY_SET,
    BINARY, BOOLEAN,
)
from pynamodb.models import Model


class AttributeTestMapAttribute(MapAttribute):
    string = UnicodeAttribute()
    binary = BinaryAttribute(legacy_encoding=False)


class AttributeTestModel(Model):

    class Meta:
        host = 'http://localhost:8000'
        table_name = 'test'

    binary_attr = BinaryAttribute(hash_key=True, legacy_encoding=False)
    binary_set_attr = BinarySetAttribute(legacy_encoding=False)
    number_attr = NumberAttribute()
    number_set_attr = NumberSetAttribute()
    unicode_attr = UnicodeAttribute()
    unicode_set_attr = UnicodeSetAttribute()
    datetime_attr = UTCDateTimeAttribute()
    bool_attr = BooleanAttribute()
    json_attr = JSONAttribute()
    map_attr = AttributeTestMapAttribute()
    raw_map_attr = MapAttribute()
    ttl_attr = TTLAttribute()
    null_attr = NullAttribute(null=True)


class CustomAttrMap(MapAttribute):
    overridden_number_attr = NumberAttribute(attr_name="number_attr")
    overridden_unicode_attr = UnicodeAttribute(attr_name="unicode_attr")


class DefaultsMap(MapAttribute):
    map_field = MapAttribute(default=dict)
    string_set_field = UnicodeSetAttribute(null=True)


class TestAttributeDescriptor:
    """
    Test Attribute Descriptors
    """
    def setup_method(self):
        self.instance = AttributeTestModel()

    def test_binary_attr(self):
        """
        Binary attribute descriptor
        """
        self.instance.binary_attr = b'test'
        assert self.instance.binary_attr == b'test'

    def test_binary_set_attr(self):
        """
        Binary set attribute descriptor
        """
        self.instance.binary_set_attr = {b'test', b'test2'}
        assert self.instance.binary_set_attr == {b'test', b'test2'}

    def test_number_attr(self):
        """
        Number attribute descriptor
        """
        self.instance.number_attr = 42
        assert self.instance.number_attr == 42

    def test_number_set_attr(self):
        """
        Number set attribute descriptor
        """
        self.instance.number_set_attr = {1, 2}
        assert self.instance.number_set_attr == {1, 2}

    def test_unicode_attr(self):
        """
        Unicode attribute descriptor
        """
        self.instance.unicode_attr = u"test"
        assert self.instance.unicode_attr == u"test"

    def test_unicode_set_attr(self):
        """
        Unicode set attribute descriptor
        """
        self.instance.unicode_set_attr = {u"test", u"test2"}
        assert self.instance.unicode_set_attr == {u"test", u"test2"}

    def test_datetime_attr(self):
        """
        Datetime attribute descriptor
        """
        now = datetime.now()
        self.instance.datetime_attr = now
        assert self.instance.datetime_attr == now

    def test_bool_attr(self):
        """
        Boolean attribute descriptor
        """
        self.instance.bool_attr = True
        assert self.instance.bool_attr is True

    def test_json_attr(self):
        """
        JSON attribute descriptor
        """
        self.instance.json_attr = {'foo': 'bar', 'bar': 42}
        assert self.instance.json_attr == {'foo': 'bar', 'bar': 42}


class TestDefault:
    def test_default(self):
        Attribute(default='test')
        Attribute(default_for_new='test')

        with pytest.raises(ValueError, match="'default' must be immutable (.*) or a callable"):
            Attribute(default=[])

        with pytest.raises(ValueError, match="'default_for_new' must be immutable (.*) or a callable"):
            Attribute(default_for_new=[])

        Attribute(default=list)
        Attribute(default_for_new=list)


class TestUTCDateTimeAttribute:
    """
    Tests UTCDateTime attributes
    """

    def setup_method(self):
        self.attr = UTCDateTimeAttribute()
        self.dt = datetime(2047, 1, 6, 8, 21, 30, 2000, tzinfo=timezone.utc)

    def test_utc_datetime_attribute(self):
        """
        UTCDateTimeAttribute.default
        """
        attr = UTCDateTimeAttribute(default=self.dt)
        assert attr.attr_type == STRING
        assert attr.default == self.dt

    def test_utc_date_time_serialize(self):
        """
        UTCDateTimeAttribute.serialize
        """
        assert self.attr.serialize(self.dt) == '2047-01-06T08:21:30.002000+0000'

    def test_utc_date_time_serialize_pre_1000(self):
        dt = self.dt.replace(year=1)
        assert self.attr.serialize(dt) == '0001-01-06T08:21:30.002000+0000'

    def test_utc_date_time_deserialize(self):
        """
        UTCDateTimeAttribute.deserialize
        """
        assert self.attr.deserialize('2047-01-06T08:21:30.002000+0000') == self.dt

    def test_utc_date_time_deserialize_pre_1000_not_padded(self):
        assert self.attr.deserialize('1-01-06T08:21:30.002000+0000') == self.dt.replace(year=1)

    @pytest.mark.parametrize(
        "invalid_string",
        [
            '2047-01-06T08:21:30.002000',       # naive datetime
            '2047-01-06T08:21:30+0000',         # missing microseconds
            '2047-01-06T08:21:30.001+0000',     # shortened microseconds
            '2047-01-06T08:21:30.002000-0000'   # "negative" utc
            '2047-01-06T08:21:30.002000+0030'   # not utc
            '2047-01-06 08:21:30.002000+0000',  # missing separator
            '2.47-01-06T08:21:30.002000+0000',
            'abcd-01-06T08:21:30.002000+0000',
            '2047-ab-06T08:21:30.002000+0000',
            '2047-01-abT08:21:30.002000+0000',
            '2047-01-06Tab:21:30.002000+0000',
            '2047-01-06T08:ab:30.002000+0000',
            '2047-01-06T08:21:ab.002000+0000',
            '2047-01-06T08:21:30.a00000+0000',
        ]
    )
    def test_utc_date_time_invalid(self, invalid_string):
        with pytest.raises(ValueError, match="does not match format"):
            self.attr.deserialize(invalid_string)


class TestBinaryAttribute:
    """
    Tests binary attributes
    """
    @pytest.mark.parametrize('legacy_encoding', [False, True])
    def test_binary_attribute(self, legacy_encoding: bool) -> None:
        attr = BinaryAttribute(legacy_encoding=legacy_encoding)
        assert attr.attr_type == BINARY
        assert attr.legacy_encoding == legacy_encoding

        attr = BinaryAttribute(default=b'foo', legacy_encoding=legacy_encoding)
        assert attr.default == b'foo'

    @pytest.mark.parametrize('legacy_encoding', [False, True])
    def test_binary_round_trip(self, legacy_encoding: bool) -> None:
        """
        BinaryAttribute round trip
        """
        attr = BinaryAttribute(legacy_encoding=legacy_encoding)
        value = b'foo'
        serial = attr.serialize(value)
        assert attr.deserialize(serial) == value

    @pytest.mark.parametrize(['legacy_encoding', 'expected'], [
        (False, b'foo'),
        (True, b'Zm9v'),
    ])
    def test_binary_serialize(self, legacy_encoding: bool, expected: bytes) -> None:
        """
        BinaryAttribute.serialize
        """
        attr = BinaryAttribute(legacy_encoding=legacy_encoding)
        assert attr.serialize(b'foo') == expected

    @pytest.mark.parametrize(['legacy_encoding', 'serialized'], [
        pytest.param(False, b'foo',id='encoded from API, decoded in _covert_binary'),
        pytest.param(True, b'Zm9v', id='double-encoded from API, once decoded in _convert_binary'),
    ])
    def test_binary_deserialize(self, legacy_encoding: bool, serialized: bytes) -> None:
        """
        BinaryAttribute.deserialize
        """
        attr = BinaryAttribute(legacy_encoding=legacy_encoding)
        assert attr.deserialize(serialized) == b'foo'


class TestBinarySetAttribute:
    @pytest.mark.parametrize(['legacy_encoding', 'expected'], [
        pytest.param(False, {b'foo', b'bar'},id='encoded from API, decoded in _covert_binary'),
        pytest.param(True, [b'Zm9v', b'YmFy'], id='double-encoded from API, once decoded in _convert_binary'),
    ])
    def test_binary_set_attribute(self, legacy_encoding: bool, expected: Set[bytes]) -> None:
        attr = BinarySetAttribute(legacy_encoding=legacy_encoding)
        assert attr.attr_type == BINARY_SET
        assert attr.legacy_encoding == legacy_encoding

    @pytest.mark.parametrize(['legacy_encoding', 'expected'], [
        (False, [b'foo', b'bar']),
        (True, [b'Zm9v', b'YmFy']),
    ])
    def test_binary_set_serialize(self, legacy_encoding: bool, expected: List[bytes]) -> None:
        """
        BinarySetAttribute.serialize
        """
        attr = BinarySetAttribute(legacy_encoding=legacy_encoding)
        assert attr.attr_type == BINARY_SET

        actual = attr.serialize({b'foo', b'bar'})
        assert isinstance(actual, list)
        assert sorted(actual) == sorted(expected)

        assert attr.serialize({}) is None

    @pytest.mark.parametrize('legacy_encoding', [False, True])
    def test_binary_set_round_trip(self, legacy_encoding: bool) -> None:
        """
        BinarySetAttribute round trip
        """
        attr = BinarySetAttribute(legacy_encoding=legacy_encoding)
        value = {b'foo', b'bar'}
        serial = attr.serialize(value)
        assert attr.deserialize(serial) == value

    @pytest.mark.parametrize(['legacy_encoding', 'serialized'], [
        pytest.param(False, [b'foo', b'bar'],id='encoded from API, decoded in _covert_binary'),
        pytest.param(True, [b'Zm9v', b'YmFy'], id='double-encoded from API, once decoded in _convert_binary'),
    ])
    def test_binary_set_deserialize(self, legacy_encoding: bool, serialized: List[bytes]) -> None:
        """
        BinarySetAttribute.deserialize
        """
        attr = BinarySetAttribute(legacy_encoding=legacy_encoding)
        assert attr.deserialize(serialized) == {b'foo', b'bar'}


class TestNumberAttribute:
    """
    Tests number attributes
    """
    def test_number_attribute(self):
        """
        NumberAttribute.default
        """
        attr = NumberAttribute()
        assert attr.attr_type == NUMBER

        attr = NumberAttribute(default=1)
        assert attr.default == 1

    def test_number_serialize(self):
        """
        NumberAttribute.serialize
        """
        attr = NumberAttribute()
        assert attr.serialize(3.141) == '3.141'
        assert attr.serialize(1) == '1'
        assert attr.serialize(12345678909876543211234234324234) == '12345678909876543211234234324234'

    def test_number_deserialize(self):
        """
        NumberAttribute.deserialize
        """
        attr = NumberAttribute()
        assert attr.deserialize('1') == 1
        assert attr.deserialize('3.141') == 3.141
        assert attr.deserialize('12345678909876543211234234324234') == 12345678909876543211234234324234

    def test_number_set_deserialize(self):
        """
        NumberSetAttribute.deserialize
        """
        attr = NumberSetAttribute()
        assert attr.attr_type == NUMBER_SET
        assert attr.deserialize([json.dumps(val) for val in sorted({1, 2})]) == {1, 2}

    def test_number_set_serialize(self):
        """
        NumberSetAttribute.serialize
        """
        attr = NumberSetAttribute()
        assert attr.serialize({1, 2}) == [json.dumps(val) for val in sorted({1, 2})]
        assert attr.serialize({}) is None

    def test_number_set_attribute(self):
        """
        NumberSetAttribute.default
        """
        attr = NumberSetAttribute(default=lambda: {1, 2})
        assert attr.default() == {1, 2}


class TestUnicodeAttribute:
    """
    Tests unicode attributes
    """
    def test_unicode_attribute(self):
        """
        UnicodeAttribute.default
        """
        attr = UnicodeAttribute()
        assert attr.attr_type == STRING

        attr = UnicodeAttribute(default='foo')
        assert attr.default == 'foo'

    def test_unicode_serialize(self):
        """
        UnicodeAttribute.serialize
        """
        attr = UnicodeAttribute()
        assert attr.serialize('foo') == 'foo'
        assert attr.serialize('') == ''
        assert attr.serialize(None) is None

    def test_unicode_deserialize(self):
        """
        UnicodeAttribute.deserialize
        """
        attr = UnicodeAttribute()
        assert attr.deserialize('foo') == 'foo'
        assert attr.deserialize(u'foo') == 'foo'
        assert attr.deserialize('') == ''
        assert attr.deserialize(None) is None

    def test_unicode_set_serialize(self):
        """
        UnicodeSetAttribute.serialize
        """
        attr = UnicodeSetAttribute()
        assert attr.attr_type == STRING_SET
        assert attr.serialize({}) is None

        expected = sorted(['foo', 'bar'])
        assert sorted(attr.serialize({'foo', 'bar'})) == expected

        expected = sorted(['True', 'False'])
        assert sorted(attr.serialize({'True', 'False'})) == expected

        expected = sorted(['true', 'false'])
        assert sorted(attr.serialize({'true', 'false'})) == expected

    def test_round_trip_unicode_set(self):
        """
        Round trip a unicode set
        """
        attr = UnicodeSetAttribute()
        orig = {'foo', 'bar'}
        assert orig == attr.deserialize(attr.serialize(orig))

        orig = {'true', 'false'}
        assert orig == attr.deserialize(attr.serialize(orig))

        orig = {'1', '2.8'}
        assert orig == attr.deserialize(attr.serialize(orig))

        orig = {'[1,2,3]', '2.8'}
        assert orig == attr.deserialize(attr.serialize(orig))

    def test_unicode_set_deserialize(self):
        """
        UnicodeSetAttribute.deserialize
        """
        attr = UnicodeSetAttribute()
        value = {'foo', 'bar'}
        assert attr.deserialize(value) == value

        value = {'True', 'False'}
        assert attr.deserialize(value) == value

        value = {'true', 'false'}
        assert attr.deserialize(value) == value

        value = {'1', '2.8'}
        assert attr.deserialize(value) == value

    def test_unicode_set_attribute(self):
        """
        UnicodeSetAttribute.default
        """
        attr = UnicodeSetAttribute()
        assert attr.attr_type == STRING_SET
        attr = UnicodeSetAttribute(default=lambda: {'foo', 'bar'})
        assert attr.default() == {'foo', 'bar'}


class TestBooleanAttribute:
    """
    Tests boolean attributes
    """
    def test_boolean_attribute(self):
        """
        BooleanAttribute.default
        """
        attr = BooleanAttribute(default=True)
        assert attr.attr_type == BOOLEAN
        assert attr.default is True

    def test_boolean_serialize(self):
        """
        BooleanAttribute.serialize
        """
        attr = BooleanAttribute()
        assert attr.serialize(True) is True
        assert attr.serialize(False) is False
        assert attr.serialize(None) is None

    def test_boolean_deserialize(self):
        """
        BooleanAttribute.deserialize
        """
        attr = BooleanAttribute()
        assert attr.deserialize(True) is True
        assert attr.deserialize(False) is False


class TestTTLAttribute:
    """
    Test TTLAttribute.
    """
    def test_default_and_default_for_new(self):
        with pytest.raises(ValueError, match='An attribute cannot have both default and default_for_new parameters'):
            TTLAttribute(default=timedelta(seconds=1), default_for_new=timedelta(seconds=2))

    @patch('time.time')
    def test_timedelta_ttl(self, mock_time):
        mock_time.side_effect = [1559692800]  # 2019-06-05 00:00:00 UTC
        model = AttributeTestModel()
        model.ttl_attr = timedelta(seconds=60)
        assert model.ttl_attr == datetime(2019, 6, 5, 0, 1, tzinfo=timezone.utc)

    def test_datetime_naive_ttl(self):
        model = AttributeTestModel()
        with pytest.raises(ValueError, match='timezone-aware'):
            model.ttl_attr = datetime(2019, 6, 5, 0, 1)
        assert model.ttl_attr is None

    def test_datetime_with_tz_ttl(self):
        model = AttributeTestModel()
        model.ttl_attr = datetime(2019, 6, 5, 0, 1, tzinfo=timezone.utc)
        assert model.ttl_attr == datetime(2019, 6, 5, 0, 1, tzinfo=timezone.utc)

    def test_ttl_attribute_wrong_type(self):
        with pytest.raises(ValueError, match='TTLAttribute value must be a timedelta or datetime'):
            model = AttributeTestModel()
            model.ttl_attr = 'wrong type'

    @patch('time.time')
    def test_serialize_timedelta(self, mock_time):
        mock_time.side_effect = [1559692800]  # 2019-06-05 00:00:00 UTC
        assert TTLAttribute().serialize(timedelta(seconds=60)) == str(1559692800 + 60)

    def test_serialize_none(self):
        model = AttributeTestModel()
        model.ttl_attr = None
        assert model.ttl_attr == None
        assert TTLAttribute().serialize(model.ttl_attr) == None

    @patch('time.time')
    def test_serialize_deserialize(self, mock_time):
        mock_time.side_effect = [1559692800, 1559692800]  # 2019-06-05 00:00:00 UTC
        model = AttributeTestModel()
        model.ttl_attr = timedelta(minutes=1)
        assert model.ttl_attr == datetime(2019, 6, 5, 0, 1, tzinfo=timezone.utc)
        s = TTLAttribute().serialize(model.ttl_attr)
        assert s == '1559692860'
        assert TTLAttribute().deserialize(s) == datetime(2019, 6, 5, 0, 1, 0, tzinfo=timezone.utc)


class TestJSONAttribute:
    """
    Tests json attributes
    """
    def test_quoted_json(self):
        attr = JSONAttribute()
        serialized = attr.serialize('\\t')
        assert attr.deserialize(serialized) == '\\t'

        serialized = attr.serialize('"')
        assert attr.deserialize(serialized) == '"'

    def test_json_attribute(self):
        """
        JSONAttribute.default
        """
        attr = JSONAttribute(default=lambda: {})
        assert attr.attr_type == STRING
        assert attr.default() == {}

    def test_json_serialize(self):
        """
        JSONAttribute.serialize
        """
        attr = JSONAttribute()
        item = {'foo': 'bar', 'bool': True, 'number': 3.141}
        assert attr.serialize(item) == json.dumps(item)
        assert attr.serialize({}) == '{}'
        assert attr.serialize(None) is None

    def test_json_deserialize(self):
        """
        JSONAttribute.deserialize
        """
        attr = JSONAttribute()
        item = {'foo': 'bar', 'bool': True, 'number': 3.141}
        encoded = json.dumps(item)
        assert attr.deserialize(encoded) == item

    def test_control_chars(self):
        """
        JSONAttribute with control chars
        """
        attr = JSONAttribute()
        item = {'foo\t': 'bar\n', 'bool': True, 'number': 3.141}
        encoded = json.dumps(item)
        assert attr.deserialize(encoded) == item


class TestMapAttribute:
    """
    Tests map with str, int, float
    """
    def test_attribute_children(self):
        person_attribute = {
            'name': 'Justin',
            'age': 12345678909876543211234234324234,
            'height': 187.96,
            'lucky_numbers': {42, 1337},
            'picture': b'\xde\xad\xbe\xef',
            'lunch': {b'\xca\xfe', b'\xf0\x0d'},
        }
        attr = MapAttribute()
        serialized = attr.serialize(person_attribute)
        assert attr.deserialize(serialized) == person_attribute

    def test_serialize_invalid_set(self):
        attr = MapAttribute()

        with pytest.raises(ValueError, match="Cannot serialize empty set"):
            attr.serialize({'test': set()})
        with pytest.raises(ValueError, match="Cannot serialize set consisting of types: <class 'int'>, <class 'str'>"):
            attr.serialize({'test': {42, "foo"}})
        with pytest.raises(ValueError, match="Cannot serialize set consisting of types: "
                                             "<class 'bytes'>, <class 'int'>"):
            attr.serialize({'test': {42, b"foo"}})

    # Special case for raw map attributes
    def test_null_attribute_raw_map(self):
        null_attribute = {
            'skip': None
        }
        attr = MapAttribute()
        serialized = attr.serialize(null_attribute)
        assert serialized == {'skip': {'NULL': True}}

    def test_null_attribute_subclassed_map(self):
        null_attribute = {
            'map_field': {},
            'string_set_field': None
        }
        attr = DefaultsMap()
        serialized = attr.serialize(null_attribute)
        assert serialized == {'map_field': {'M': {}}}

    def test_null_attribute_map_after_serialization(self):
        null_attribute = {
            'map_field': {},
            'string_set_field': {},
        }
        attr = DefaultsMap()
        serialized = attr.serialize(null_attribute)
        assert serialized == {'map_field': {'M': {}}}

    def test_map_of_map(self):
        attribute = {
            'name': 'Justin',
            'metrics': {
                'age': 31,
                'height': 187.96
            }
        }
        attr = MapAttribute()
        serialized = attr.serialize(attribute)
        assert attr.deserialize(serialized) == attribute

    def test_map_overridden_attrs_accessors(self):
        attr = CustomAttrMap(**{
            'overridden_number_attr': 10,
            'overridden_unicode_attr': "Hello"
        })

        assert attr.overridden_number_attr == 10
        assert attr.overridden_unicode_attr == "Hello"

    def test_map_overridden_attrs_serialize(self):
        attribute = {
            'overridden_number_attr': 10,
            'overridden_unicode_attr': "Hello"
        }
        expected = {'number_attr': {'N': '10'}, 'unicode_attr': {'S': 'Hello'}}
        assert CustomAttrMap().serialize(attribute) == expected

    def test_additional_attrs_deserialize(self):
        raw_data = {
            'number_attr': {
                'N': '10'},
            'unicode_attr': {
                'S': 'Hello'
            },
            'undeclared_attr': {
                'S': 'Goodbye'
            }
        }
        expected = {
            'overridden_number_attr': 10,
            'overridden_unicode_attr': "Hello"
        }
        assert CustomAttrMap().deserialize(raw_data).attribute_values == expected

    def test_defaults(self):
        item = DefaultsMap()
        assert item.validate()
        assert DefaultsMap().serialize(item) == {
            'map_field': {
                'M': {}
            }
        }

    def test_raw_set_attr(self):
        item = AttributeTestModel()
        item.raw_map_attr = {}
        item.raw_map_attr.foo = 'bar'
        item.raw_map_attr.num = 3
        item.raw_map_attr.nested = {'nestedfoo': 'nestedbar'}

        assert item.raw_map_attr['foo'] == 'bar'
        assert item.raw_map_attr['num'] == 3
        assert item.raw_map_attr['nested']['nestedfoo'] == 'nestedbar'

    def test_raw_set_item(self):
        item = AttributeTestModel()
        item.raw_map_attr = {}
        item.raw_map_attr['foo'] = 'bar'
        item.raw_map_attr['num'] = 3
        item.raw_map_attr['nested'] = {'nestedfoo': 'nestedbar'}

        assert item.raw_map_attr['foo'] == 'bar'
        assert item.raw_map_attr['num'] == 3
        assert item.raw_map_attr['nested']['nestedfoo'] == 'nestedbar'

    def test_raw_map_from_dict(self):
        item = AttributeTestModel(
            raw_map_attr={
                "foo": "bar",
                "num": 3,
                "nested": {
                    "nestedfoo": "nestedbar"
                }
            }
        )

        assert item.raw_map_attr['foo'] == 'bar'
        assert item.raw_map_attr['num'] == 3

    def test_raw_map_access(self):
        raw = {
            "foo": "bar",
            "num": 3,
            "nested": {
                "nestedfoo": "nestedbar"
            }
        }
        attr = MapAttribute(**raw)

        for k, v in raw.items():
            assert attr[k] == v

    def test_raw_map_iter(self):
        raw = {
            "foo": "bar",
            "num": 3,
            "nested": {
                "nestedfoo": "nestedbar"
            }
        }
        attr = MapAttribute(**raw)

        assert sorted(iter(raw)) == sorted(iter(attr))

    def test_raw_map_json_serialize(self):
        raw = {
            "foo": "bar",
            "num": 3,
            "nested": {
                "nestedfoo": "nestedbar"
            }
        }

        serialized_raw = json.dumps(raw, sort_keys=True)
        serialized_attr_from_raw = json.dumps(
            AttributeTestModel(raw_map_attr=raw).raw_map_attr.as_dict(),
            sort_keys=True)
        serialized_attr_from_map = json.dumps(
            AttributeTestModel(raw_map_attr=MapAttribute(**raw)).raw_map_attr.as_dict(),
            sort_keys=True)

        assert serialized_attr_from_raw == serialized_raw
        assert serialized_attr_from_map == serialized_raw

    def test_typed_and_raw_map_json_serialize(self):
        class TypedMap(MapAttribute):
            map_attr = MapAttribute()

        class SomeModel(Model):
            key = NumberAttribute(hash_key=True)
            typed_map = TypedMap()

        item = SomeModel(
            typed_map=TypedMap(map_attr={'foo': 'bar'})
        )

        assert json.dumps({'map_attr': {'foo': 'bar'}}) == json.dumps(item.typed_map.as_dict())

    def test_json_serialize(self):
        class JSONMapAttribute(MapAttribute):
            arbitrary_data = JSONAttribute()

            def __eq__(self, other):
                return self.arbitrary_data == other.arbitrary_data

        item = {'foo': 'bar', 'bool': True, 'number': 3.141}
        json_map = JSONMapAttribute(arbitrary_data=item)
        serialized = json_map.serialize(json_map)
        deserialized = json_map.deserialize(serialized)
        assert isinstance(deserialized, JSONMapAttribute)
        assert deserialized == json_map
        assert deserialized.arbitrary_data == item

    def test_serialize_datetime(self):
        class CustomMapAttribute(MapAttribute):
            date_attr = UTCDateTimeAttribute()

        cm = CustomMapAttribute(date_attr=datetime(2017, 1, 1))
        serialized_datetime = cm.serialize(cm)
        expected_serialized_value = {
            'date_attr': {
                'S': u'2017-01-01T00:00:00.000000+0000'
            }
        }
        assert serialized_datetime == expected_serialized_value

    def test_complex_map_accessors(self):
        class NestedThing(MapAttribute):
            double_nested = MapAttribute()
            double_nested_renamed = MapAttribute(attr_name='something_else')

        class ThingModel(Model):
            key = NumberAttribute(hash_key=True)
            nested = NestedThing()

        t = ThingModel(nested=NestedThing(
            double_nested={'hello': 'world'},
            double_nested_renamed={'foo': 'bar'})
        )

        assert t.nested.double_nested.as_dict() == {'hello': 'world'}
        assert t.nested.double_nested_renamed.as_dict() == {'foo': 'bar'}
        assert t.nested.double_nested.hello == 'world'
        assert t.nested.double_nested_renamed.foo == 'bar'
        assert t.nested['double_nested'].as_dict() == {'hello': 'world'}
        assert t.nested['double_nested_renamed'].as_dict() == {'foo': 'bar'}
        assert t.nested['double_nested']['hello'] == 'world'
        assert t.nested['double_nested_renamed']['foo'] == 'bar'

        with pytest.raises(AttributeError):
            bad = t.nested.double_nested.bad
        with pytest.raises(AttributeError):
            bad = t.nested.bad
        with pytest.raises(AttributeError):
            bad = t.nested.something_else
        with pytest.raises(KeyError):
            bad = t.nested.double_nested['bad']
        with pytest.raises(KeyError):
            bad = t.nested['something_else']

    def test_attribute_paths_subclassing(self):
        class SubMapAttribute(MapAttribute):
            foo = UnicodeAttribute(attr_name='dyn_foo')

        class SubSubMapAttribute(SubMapAttribute):
            bar = UnicodeAttribute(attr_name='dyn_bar')

        class SubModel(Model):
            key = NumberAttribute(hash_key=True)
            sub_map = SubMapAttribute(attr_name='dyn_sub_map')

        class SubSubModel(SubModel):
            sub_sub_map = SubSubMapAttribute()

        assert SubModel.sub_map.foo.attr_name == 'dyn_foo'
        assert SubModel.sub_map.foo.attr_path == ['dyn_sub_map', 'dyn_foo']
        assert SubSubModel.sub_map.foo.attr_name == 'dyn_foo'
        assert SubSubModel.sub_map.foo.attr_path == ['dyn_sub_map', 'dyn_foo']
        assert SubSubModel.sub_sub_map.foo.attr_name == 'dyn_foo'
        assert SubSubModel.sub_sub_map.foo.attr_path == ['sub_sub_map', 'dyn_foo']
        assert SubSubModel.sub_sub_map.bar.attr_name == 'dyn_bar'
        assert SubSubModel.sub_sub_map.bar.attr_path == ['sub_sub_map', 'dyn_bar']

    def test_attribute_paths_wrapping(self):
        class InnerMapAttribute(MapAttribute):
            map_attr = MapAttribute(attr_name='dyn_map_attr')

        class MiddleMapAttributeA(MapAttribute):
            inner_map = InnerMapAttribute(attr_name='dyn_in_map_a')

        class MiddleMapAttributeB(MapAttribute):
            inner_map = InnerMapAttribute(attr_name='dyn_in_map_b')

        class OuterMapAttribute(MapAttribute):
            mid_map_a = MiddleMapAttributeA()
            mid_map_b = MiddleMapAttributeB()

        class MyModel(Model):
            key = NumberAttribute(hash_key=True)
            outer_map = OuterMapAttribute(attr_name='dyn_out_map')

        mid_map_a_map_attr = MyModel.outer_map.mid_map_a.inner_map.map_attr
        mid_map_b_map_attr = MyModel.outer_map.mid_map_b.inner_map.map_attr

        assert mid_map_a_map_attr.attr_name == 'dyn_map_attr'
        assert mid_map_a_map_attr.attr_path == ['dyn_out_map', 'mid_map_a', 'dyn_in_map_a', 'dyn_map_attr']
        assert mid_map_b_map_attr.attr_name == 'dyn_map_attr'
        assert mid_map_b_map_attr.attr_path == ['dyn_out_map', 'mid_map_b', 'dyn_in_map_b', 'dyn_map_attr']

    def test_required_elements(self):
        class InnerMapAttribute(MapAttribute):
            foo = UnicodeAttribute()

        class OuterMapAttribute(MapAttribute):
            inner_map = InnerMapAttribute()

        outer_map_attribute = OuterMapAttribute()
        with pytest.raises(ValueError):
            outer_map_attribute.serialize(outer_map_attribute)

        outer_map_attribute = OuterMapAttribute(inner_map={})
        with pytest.raises(ValueError):
            outer_map_attribute.serialize(outer_map_attribute)

        outer_map_attribute = OuterMapAttribute(inner_map=MapAttribute())
        with pytest.raises(ValueError):
            outer_map_attribute.serialize(outer_map_attribute)

        outer_map_attribute = OuterMapAttribute(inner_map={'foo': 'bar'})
        serialized = outer_map_attribute.serialize(outer_map_attribute)
        assert serialized == {'inner_map': {'M': {'foo': {'S': 'bar'}}}}


class TestDynamicMapAttribute:

    class CreatedAtTestModel(Model):
        class CreatedAtMap(DynamicMapAttribute):
            created_at = UTCDateTimeAttribute()
        test_map = CreatedAtMap(default=dict)

    def test_serialize(self):
        test_model = TestDynamicMapAttribute.CreatedAtTestModel()
        test_model.test_map.created_at = datetime(2017, 1, 1, tzinfo=timezone.utc)
        test_model.test_map.foo = 'bar'
        test_model.test_map.empty = None
        assert test_model.serialize() == {'test_map': {'M': {
            'created_at': {'S': '2017-01-01T00:00:00.000000+0000'},
            'foo': {'S': 'bar'},
            'empty': {'NULL': True},
        }}}

    def test_deserialize(self):
        serialized = {'test_map': {'M': {
            'created_at': {'S': '2017-01-01T00:00:00.000000+0000'},
            'foo': {'S': 'bar'},
            'empty': {'NULL': True},
        }}}
        test_model = TestDynamicMapAttribute.CreatedAtTestModel()
        test_model.deserialize(serialized)
        assert test_model.test_map.created_at == datetime(2017, 1, 1, tzinfo=timezone.utc)
        assert test_model.test_map.foo == 'bar'
        assert test_model.test_map.empty is None


class TestListAttribute:

    def test_roundtrip_untyped(self) -> None:
        string_list_attribute = ListAttribute()
        values = [
            None,  # NULL
            'foo',  # S
            '',  # S
            42,  # NUMBER
            True,  # BOOL
            b'foo',  # B
            {42, 43},  # NS of ints
            {42.5, 43.5},  # NS of floats
            {42, 43.5},  # NS of ints and floats
            {'foo', 'bar'},  # SS
            {b'foo', b'bar'},  # BS
            {'foo': 'bar'},  # M with one S
            ['foo', 'bar'],  # L of S
        ]
        serialized = string_list_attribute.serialize(values)
        assert string_list_attribute.deserialize(serialized) == values

    def test_serialize_invalid_set(self) -> None:
        string_list_attribute = ListAttribute()
        with pytest.raises(ValueError, match="Cannot serialize empty set"):
            string_list_attribute.serialize([set()])
        with pytest.raises(ValueError, match="Cannot serialize set consisting of types: <class 'int'>, <class 'str'>"):
            string_list_attribute.serialize([{42, "foo"}])
        with pytest.raises(ValueError, match="Cannot serialize set consisting of types: "
                                             "<class 'bytes'>, <class 'int'>"):
            string_list_attribute.serialize([{42, b"foo"}])

    def test_list_of_binary_set(self) -> None:
        string_list_attribute = ListAttribute(of=BinarySetAttribute)
        values = [{b'bar', b'baz'}]
        serialized = string_list_attribute.serialize(values)
        assert string_list_attribute.deserialize(serialized) == values

    def test_list_type_error(self):
        string_list_attribute = ListAttribute(of=UnicodeAttribute)

        with pytest.raises(ValueError, match='List elements must be of type: UnicodeAttribute'):
            string_list_attribute.serialize([MapAttribute(foo='bar')])

        with pytest.raises(TypeError, match=r"Cannot deserialize '\[0\]' attribute from type: S"):
            string_list_attribute.deserialize([{'M': {'foo': {'S': 'bar'}}}])

    def test_serialize_null(self):
        string_set_list_attribute = ListAttribute(of=UnicodeSetAttribute)
        list_with_empty_set = [{'foo'}, {}, None]
        serialized = string_set_list_attribute.serialize(list_with_empty_set)
        assert string_set_list_attribute.deserialize(serialized) == [{'foo'}, None, None]


class TestMapAndListAttribute:

    def test_map_of_list(self):
        grocery_list = {
            'fruit': ['apple', 'pear', 32],
            'veggies': ['broccoli', 'potatoes', 5]
        }
        serialized = MapAttribute().serialize(grocery_list)
        assert MapAttribute().deserialize(serialized) == grocery_list

    def test_map_of_list_of_map(self):
        family_attributes = {
            'phillips': [
                {
                    'name': 'Justin',
                    'age': 31,
                    'height': 187.96
                },
                {
                    'name': 'Galen',
                    'age': 29,
                    'height': 193.04,
                    'male': True
                },
                {
                    'name': 'Colin',
                    'age': 32,
                    'height': 203.2,
                    'male': True,
                    'hasChild': True
                }
            ],
            'workman': [
                {
                    'name': 'Mandy',
                    'age': 29,
                    'height': 157.48,
                    'female': True
                },
                {
                    'name': 'Rodney',
                    'age': 31,
                    'height': 175.26,
                    'hasChild': False
                }
            ]
        }
        serialized = MapAttribute().serialize(family_attributes)
        assert MapAttribute().deserialize(serialized) == family_attributes

    def test_list_of_map_with_of(self):
        class Person(MapAttribute):
            name = UnicodeAttribute()
            age = NumberAttribute()

            def __lt__(self, other):
                return self.name < other.name

            def __eq__(self, other):
                return (self.name == other.name and
                        self.age == other.age)

        person1 = Person()
        person1.name = 'john'
        person1.age = 40

        person2 = Person()
        person2.name = 'Dana'
        person2.age = 41

        inp = [person1, person2]

        list_attribute = ListAttribute(default=list, of=Person)
        serialized = list_attribute.serialize(inp)
        deserialized = list_attribute.deserialize(serialized)
        assert sorted(deserialized) == sorted(inp)

    def test_list_of_map_with_of_and_custom_attribute(self, mocker):

        class CustomMapAttribute(MapAttribute):
            custom = NumberAttribute()

            def __eq__(self, other):
                return self.custom == other.custom

        serialize_mock = mocker.spy(CustomMapAttribute.custom, 'serialize',)
        deserialize_mock = mocker.spy(CustomMapAttribute.custom, 'deserialize')

        attribute1 = CustomMapAttribute()
        attribute1.custom = 1

        attribute2 = CustomMapAttribute()
        attribute2.custom = 2

        inp = [attribute1, attribute2]

        list_attribute = ListAttribute(default=list, of=CustomMapAttribute)
        serialized = list_attribute.serialize(inp)
        deserialized = list_attribute.deserialize(serialized)

        assert deserialized == inp
        assert serialize_mock.call_args_list == [call(1), call(2)]
        assert deserialize_mock.call_args_list == [call('1'), call('2')]


class TestVersionAttribute:
    def test_serialize(self):
        attr = VersionAttribute()
        assert attr.attr_type == NUMBER
        assert attr.serialize(3.141) == '3'
        assert attr.serialize(1) == '1'
        assert attr.serialize(12345678909876543211234234324234) == '12345678909876543211234234324234'

    def test_deserialize(self):
        attr = VersionAttribute()
        assert attr.deserialize('1') == 1
        assert attr.deserialize('3.141') == 3
        assert attr.deserialize('12345678909876543211234234324234') == 12345678909876543211234234324234


class DictTestMapAttribute(MapAttribute):
    string = UnicodeAttribute(null=True)
    binary = BinaryAttribute(null=True, legacy_encoding=False)


class DictTestModel(Model):
    binary_attr = BinaryAttribute(null=True, legacy_encoding=False)
    binary_set_attr = BinarySetAttribute(null=True, legacy_encoding=False)
    number_attr = NumberAttribute(null=True)
    number_set_attr = NumberSetAttribute(null=True)
    unicode_attr = UnicodeAttribute(null=True)
    unicode_set_attr = UnicodeSetAttribute(null=True)
    datetime_attr = UTCDateTimeAttribute(null=True)
    bool_attr = BooleanAttribute(null=True)
    json_attr = JSONAttribute(null=True)
    map_attr = DictTestMapAttribute(null=True)
    raw_map_attr = MapAttribute(null=True)
    ttl_attr = TTLAttribute(null=True)
    null_attr = NullAttribute(null=True)
    list_attr = ListAttribute(of=DictTestMapAttribute, null=True)


class TestNormalDict:
    def test_to_simple_dict(self):
        dt = datetime(2022, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
        test_model = DictTestModel()
        test_model.number_attr = 1
        test_model.unicode_attr = 'foo'
        test_model.datetime_attr = dt
        test_model.bool_attr = True
        test_model.json_attr = {'foo': 'bar'}
        test_model.raw_map_attr = {
            'string': 'bar',
        }
        test_model.map_attr = DictTestMapAttribute(
            string='bar',
        )
        test_model.ttl_attr = dt
        test_model.null_attr = True
        test_model.list_attr = [
            DictTestMapAttribute(string='bar'),
        ]

        actual = test_model.to_simple_dict()

        assert actual == {
            'bool_attr': True,
            'datetime_attr': '2022-12-31T23:59:59.000000+0000',
            'json_attr': '{"foo": "bar"}',
            'list_attr': [
                {
                    'string': 'bar',
                },
            ],
            'map_attr': {
                'string': 'bar',
            },
            'null_attr': None,
            'number_attr': 1,
            'raw_map_attr': {
                'string': 'bar',
            },
            'ttl_attr': 1672531199,
            'unicode_attr': 'foo',
        }

        # ensure it JSON--serializable
        _ = json.dumps(actual)

    def test_to_simple_dict__not_force(self):
        with pytest.raises(ValueError):
            DictTestModel(binary_attr=b'foo').to_simple_dict(force=False)

        with pytest.raises(ValueError):
            DictTestModel(number_set_attr={1, 2, 3}).to_simple_dict(force=False)

        with pytest.raises(ValueError):
            DictTestModel(unicode_set_attr={'foo', 'bar'}).to_simple_dict(force=False)

        with pytest.raises(ValueError):
            DictTestModel(binary_set_attr={b'foo', b'bar'}).to_simple_dict(force=False)

        with pytest.raises(ValueError):
            DictTestModel(raw_map_attr={'string': 'bar','binary': b'baz'}).to_simple_dict(force=False)

        with pytest.raises(ValueError):
            DictTestModel(DictTestMapAttribute(string='bar', binary=b'foo')).to_simple_dict(force=False)

    def test_to_simple_dict__force(self):
        test_model = DictTestModel()
        test_model.binary_attr = b'foo'
        test_model.binary_set_attr = [b'foo', b'bar']
        test_model.number_set_attr = [1, 2, 3]
        test_model.unicode_set_attr = ['foo', 'bar']
        test_model.raw_map_attr = {
            'string': 'bar',
            'binary': b'baz',  # this will not roundtrip :(
        }
        test_model.map_attr = DictTestMapAttribute(
            string='bar',
            binary=b'foo',
        )

        actual = test_model.to_simple_dict(force=True)

        assert actual == {
            'binary_attr': 'Zm9v',
            'binary_set_attr': ['Zm9v', 'YmFy'],
            'map_attr': {
                'binary': 'Zm9v',
                'string': 'bar',
            },
            'number_set_attr': [1, 2, 3],
            'raw_map_attr': {
                'binary': 'YmF6',
                'string': 'bar',
            },
            'unicode_set_attr': ['foo', 'bar'],
        }

        # ensure it JSON--serializable
        _ = json.dumps(actual)

    def test_from_simple_dict(self):
        simple_dict = {
            'binary_attr': 'Zm9v',
            'binary_set_attr': ['Zm9v', 'YmFy'],
            'bool_attr': True,
            'datetime_attr': '2022-12-31T23:59:59.000000+0000',
            'json_attr': '{"foo": "bar"}',
            'map_attr': {
                'binary': 'YmF6',
                'string': 'bar',
            },
            'raw_map_attr': {
                'binary': 'YmF6',
                'string': 'bar',
            },
            'null_attr': None,
            'number_attr': 1,
            'number_set_attr': [1, 2, 3],
            'ttl_attr': 1672531199,
            'unicode_attr': 'foo',
            'unicode_set_attr': ['baz'],
        }

        test_model = DictTestModel()
        test_model.from_simple_dict(simple_dict)

        expected_dt = datetime(2022, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
        assert test_model.binary_attr == b'foo'
        assert test_model.binary_set_attr == {b'foo', b'bar'}
        assert test_model.number_attr == 1
        assert test_model.number_set_attr == {1, 2, 3}
        assert test_model.unicode_attr == 'foo'
        assert test_model.unicode_set_attr == {'baz'}
        assert test_model.datetime_attr == expected_dt
        assert test_model.bool_attr is True
        assert test_model.json_attr == {'foo': 'bar'}
        assert test_model.map_attr.binary == b'baz'
        assert test_model.map_attr.string == 'bar'
        assert test_model.raw_map_attr.binary == 'YmF6'  # this did not round-trip :(
        assert test_model.raw_map_attr.string == 'bar'
        assert test_model.ttl_attr == expected_dt
        assert test_model.null_attr is None


class TestDynamoDBDict:
    def test_to_dynamodb_dict(self):
        dt = datetime(2022, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
        test_model = DictTestModel()
        test_model.number_attr = 1
        test_model.unicode_attr = 'foo'
        test_model.datetime_attr = dt
        test_model.bool_attr = True
        test_model.json_attr = {'foo': 'bar'}
        test_model.raw_map_attr = {
            'binary': b'foo',
            'string': 'bar',
        }
        test_model.map_attr = DictTestMapAttribute(
            binary=b'foo',
            string='bar',
        )
        test_model.list_attr = [
            DictTestMapAttribute(
                binary=b'foo',
                string='bar',
            )
        ]
        test_model.ttl_attr = dt
        test_model.null_attr = True
        test_model.binary_attr = b'foo'
        test_model.binary_set_attr = [b'foo', b'bar']
        test_model.number_set_attr = [1, 2, 3]
        test_model.unicode_set_attr = ['foo', 'bar']

        actual = test_model.to_dynamodb_dict()

        assert actual == {
            'binary_attr': {
                'B': 'Zm9v'
            },
            'binary_set_attr': {
                'BS': ['Zm9v', 'YmFy']
            },
            'bool_attr': {
                'BOOL': True
            },
            'datetime_attr': {
                'S': '2022-12-31T23:59:59.000000+0000'
            },
            'json_attr': {
                'S': '{"foo": "bar"}'
            },
            'list_attr': {
                'L': [
                    {
                        'M': {
                            'binary': {
                                'B': 'Zm9v'
                            },
                            'string': {
                                'S': 'bar'
                            }
                        }
                    },
                ],
            },
            'map_attr': {
                'M': {
                    'binary': {
                        'B': 'Zm9v'
                    },
                    'string': {
                        'S': 'bar'
                    }
                }
            },
            'null_attr': {
                'NULL': True
            },
            'number_attr': {
                'N': '1'
            },
            'number_set_attr': {
                'NS': ['1', '2', '3']
            },
            'raw_map_attr': {
                'M': {
                    'binary': {
                        'B': 'Zm9v'
                    },
                    'string': {
                        'S': 'bar'
                    }
                }
            },
            'ttl_attr': {
                'N': '1672531199'
            },
            'unicode_attr': {
                'S': 'foo'
            },
            'unicode_set_attr': {
                'SS': ['foo', 'bar']
            }
        }

        # ensure it JSON--serializable
        _ = json.dumps(actual)

    def test_from_dynamodb_dict(self):
        dynamodb_dict = {
            'binary_attr': {
                'B': 'Zm9v'
            },
            'binary_set_attr': {
                'BS': ['Zm9v', 'YmFy']
            },
            'bool_attr': {
                'BOOL': True
            },
            'datetime_attr': {
                'S': '2022-12-31T23:59:59.000000+0000'
            },
            'json_attr': {
                'S': '{"foo": "bar"}'
            },
            'map_attr': {
                'M': {
                    'binary': {
                        'B': 'Zm9v'
                    },
                    'string': {
                        'S': 'bar'
                    }
                }
            },
            'null_attr': {
                'NULL': True
            },
            'number_attr': {
                'N': '1'
            },
            'number_set_attr': {
                'NS': ['1', '2', '3']
            },
            'raw_map_attr': {
                'M': {
                    'binary': {
                        'B': 'Zm9v'
                    },
                    'string': {
                        'S': 'bar'
                    }
                }
            },
            'ttl_attr': {
                'N': '1672531199'
            },
            'unicode_attr': {
                'S': 'foo'
            },
            'unicode_set_attr': {
                'SS': ['foo', 'bar']
            }
        }

        test_model = DictTestModel()
        test_model.from_dynamodb_dict(dynamodb_dict)

        expected_dt = datetime(2022, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
        assert test_model.binary_attr == b'foo'
        assert test_model.binary_set_attr == {b'foo', b'bar'}
        assert test_model.number_attr == 1
        assert test_model.number_set_attr == {1, 2, 3}
        assert test_model.unicode_attr == 'foo'
        assert test_model.unicode_set_attr == {'foo', 'bar'}
        assert test_model.datetime_attr == expected_dt
        assert test_model.bool_attr is True
        assert test_model.json_attr == {'foo': 'bar'}
        assert test_model.map_attr.binary == b'foo'
        assert test_model.map_attr.string == 'bar'
        assert test_model.raw_map_attr.binary == b'foo'
        assert test_model.raw_map_attr.string == 'bar'
        assert test_model.ttl_attr == expected_dt
        assert test_model.null_attr is None



from amazondax import AmazonDaxClient


OP_WRITE = {
    'PutItem': 'put_item',
    'DeleteItem': 'delete_item',
    'UpdateItem': 'update_item',
    'BatchWriteItem': 'batch_write_item',
    'TransactWriteItems': 'transact_write_items',

}

OP_READ = {
    'GetItem': 'get_item',
    'BatchGetItem': 'batch_get_item',
    # query and scan has a serious consistency issue
    # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/DAX.consistency.html#DAX.consistency.query-cache
    'Query': 'query',
    'Scan': 'scan',
    'TransactGetItems': 'transact_get_items',
}

OP_NAME_TO_METHOD = OP_WRITE.copy()
OP_NAME_TO_METHOD.update(OP_READ)


class DaxClient(object):

    def __init__(self, endpoints, region_name):
        self.connection = AmazonDaxClient(
            endpoints=endpoints,
            region_name=region_name
        )

    def dispatch(self, operation_name, kwargs):
        method = getattr(self.connection, OP_NAME_TO_METHOD[operation_name])
        return method(**kwargs)
